import typing

import numpy as np

from libertem.common.math import flat_nonzero
from .file import File
from .utils import FileTree
from .tiling import default_get_read_ranges
from .tiling_scheme import TilingScheme


class FileSet:
    """
    Parameters
    ----------
    files
        files that are part of a partition or dataset
    """
    def __init__(
        self, files: list[File],
        frame_header_bytes: int = 0, frame_footer_bytes: int = 0,
    ):
        self._files = files
        assert len(files) > 0
        self._tree = FileTree.make(files)
        if self._tree is None:
            raise ValueError(str(files))
        # FIXME: maybe should be moved into the array representation
        # if the fileset, taken from individual files
        self._frame_header_bytes = frame_header_bytes
        self._frame_footer_bytes = frame_footer_bytes

    def get_for_range(self, start, stop):
        """
        return new FileSet filtered for files having frames in the [start, stop) range
        """
        files = self._get_files_for_range(start, stop)
        return self._clone(
            files=files,
            frame_header_bytes=self._frame_header_bytes,
            frame_footer_bytes=self._frame_footer_bytes,
        )

    def _clone(self, *args, **kwargs):
        return self.__class__(*args, **kwargs)

    def _get_files_for_range(self, start, stop):
        """
        return new list of files filtered for files having frames in the [start, stop) range
        """
        files = []
        for f in self.files_from(start):
            if f.start_idx > stop:
                break
            files.append(f)
        assert len(files) > 0
        return files

    def files_from(self, start):
        lower_bound, f = self._tree.search_start(start)
        for idx in range(lower_bound, len(self._files)):
            yield self._files[idx]

    def __iter__(self):
        return iter(self._files)

    def __len__(self):
        return len(self._files)

    def __getitem__(self, idx):
        return self._files[idx]

    def __repr__(self):
        return f"<{self.__class__.__name__} {self._files!r}>"

    def get_as_arr(self):
        fileset_arr = np.zeros((len(self), 4), dtype=np.int64)
        for idx, f in enumerate(self._files):
            fileset_arr[idx] = (f.start_idx, f.end_idx, idx, f.file_header_bytes)
        return fileset_arr

    def get_read_ranges(
        self, start_at_frame: int, stop_before_frame: int,
        dtype, tiling_scheme: TilingScheme, sync_offset: int = 0,
        roi: typing.Union[np.ndarray, None] = None,
    ):
        fileset_arr = self.get_as_arr()
        roi_nonzero = None
        if roi is not None:
            roi_nonzero = flat_nonzero(roi).astype(np.int64)
        return default_get_read_ranges(
            start_at_frame=start_at_frame,
            stop_before_frame=stop_before_frame,
            roi_nonzero=roi_nonzero,
            depth=tiling_scheme.depth,
            slices_arr=tiling_scheme.slices_array,
            fileset_arr=fileset_arr,
            sig_shape=tuple(tiling_scheme.dataset_shape.sig),
            sync_offset=sync_offset,
            bpp=np.dtype(dtype).itemsize * 8,
            frame_header_bytes=self._frame_header_bytes,
            frame_footer_bytes=self._frame_footer_bytes,
        )



import datetime
import time
import os
from contextlib import contextmanager

import numpy as np
import sparse
import pytest

from sparseconverter import (
    DENSE_BACKENDS, ND_BACKENDS, NUMPY, SPARSE_BACKENDS, SPARSE_GCXS, ArrayBackend, for_backend
)
from libertem.common.sparse import to_dense
from libertem.udf import UDF
import libertem.common.backend as bae
from libertem.udf.raw import PickUDF
from libertem.udf.masks import ApplyMasksUDF
from libertem.io.corrections import CorrectionSet
from libertem.io.corrections.detector import correct
from libertem.io.dataset.base.backend import IOBackend, IOBackendImpl
from libertem.common.backend import get_use_cpu, get_use_cuda, set_use_cpu, set_use_cuda
from libertem.utils.devices import detect


def _naive_mask_apply(masks, data):
    """
    masks: list of masks
    data: 4d array of input data

    returns array of shape (num_masks, scan_y, scan_x)
    """
    data = for_backend(data, NUMPY)
    assert len(data.shape) == 4
    for mask in masks:
        assert mask.shape == data.shape[2:], "mask doesn't fit frame size"

    dtype = np.result_type(*(m.dtype for m in masks), data.dtype)
    res = np.zeros((len(masks),) + tuple(data.shape[:2]), dtype=dtype)
    for n in range(len(masks)):
        mask = to_dense(masks[n])
        for i in range(data.shape[0]):
            for j in range(data.shape[1]):
                item = data[i, j].ravel().dot(mask.ravel())
                res[n, i, j] = item
    return res


# This function introduces asymmetries so that errors won't average out so
# easily with large data sets
def _mk_random(size, dtype='float32', array_backend=NUMPY, sparse_density=None):
    size = tuple(size)
    if array_backend not in ND_BACKENDS and len(size) != 2:
        raise ValueError(f"Format {array_backend} does not support size {size}")
    dtype = np.dtype(dtype)
    if array_backend in SPARSE_BACKENDS:
        if array_backend == SPARSE_GCXS:
            form = 'gcxs'
        else:
            form = 'coo'
        data = for_backend(
            sparse.random(
                size, format=form, density=sparse_density,
            ).astype(dtype),
            array_backend,
        )
    elif array_backend in DENSE_BACKENDS:
        if dtype.kind == 'c':
            choice = [0, 1, -1, 0+1j, 0-1j, 2.3+17j, -23+42j]
        else:
            choice = [0, 1]
        data = np.random.choice(choice, size=size).astype(dtype)
        coords2 = tuple(np.random.choice(range(c)) for c in size)
        coords10 = tuple(np.random.choice(range(c)) for c in size)
        data[coords2] = np.random.choice(choice) * sum(size)
        data[coords10] = np.random.choice(choice) * 10 * sum(size)
        data = for_backend(data, array_backend)
    else:
        raise ValueError(f"Don't understand array format {array_backend}.")
    if data.dtype != dtype:
        raise ValueError(f"Can't make array with format {array_backend} and dtype {dtype}.")
    if data.shape != size:
        raise ValueError(
            f"Can't make array with format {array_backend} and shape {size}, "
            f"got shape {data.shape}."
        )
    return data


def assert_msg(msg, msg_type, status='ok'):
    print(time.time(), datetime.datetime.now(), msg, msg_type, status)
    assert msg['status'] == status
    assert msg['messageType'] == msg_type, \
        "expected: {}, is: {}".format(msg_type, msg['messageType'])


class PixelsumUDF(UDF):
    def get_result_buffers(self):
        return {
            'pixelsum': self.buffer(
                kind="nav", dtype="float32"
            )
        }

    def process_frame(self, frame):
        assert frame.shape == (16, 16)
        assert self.results.pixelsum.shape == (1,)
        self.results.pixelsum[:] = np.sum(frame)


class MockFile:
    def __init__(self, start_idx, end_idx):
        self.start_idx = start_idx
        self.end_idx = end_idx

    def __repr__(self):
        return "<MockFile: [%d, %d)>" % (self.start_idx, self.end_idx)


class DebugDeviceUDF(UDF):
    def __init__(self, backends=None):
        if backends is None:
            backends = ('cupy', 'numpy')
        super().__init__(backends=backends)

    def get_result_buffers(self):
        return {
            'device_id': self.buffer(kind="single", dtype="object"),
            'on_device': self.buffer(kind="sig", dtype=np.float32, where="device"),
            'device_class': self.buffer(kind="nav", dtype="object"),
            'backend': self.buffer(kind="single", dtype="object"),
        }

    def preprocess(self):
        self.results.device_id[0] = dict()
        self.results.backend[0] = dict()

    def process_partition(self, partition):
        cpu = bae.get_use_cpu()
        cuda = bae.get_use_cuda()
        self.results.device_id[0][self.meta.slice] = {
            "cpu": cpu,
            "cuda": cuda
        }
        self.results.on_device[:] += self.xp.sum(partition, axis=0)
        self.results.device_class[:] = self.meta.device_class
        self.results.backend[0][self.meta.slice] = str(self.xp)
        print(f"meta device_class {self.meta.device_class}")

    def merge(self, dest, src):
        de, sr = dest.device_id[0], src.device_id[0]
        for key, value in sr.items():
            assert key not in de
            de[key] = value

        de, sr = dest.backend[0], src.backend[0]
        for key, value in sr.items():
            assert key not in de
            de[key] = value

        dest.on_device[:] += src.on_device
        dest.device_class[:] = src.device_class

    def get_backends(self):
        return self.params.backends


class ValidationUDF(UDF):
    '''
    UDF to compare a dataset against an array-like
    object using a validation function, by default np.allclose().

    Note that the reference should have a flattened nav and the ROI already
    applied since it is compared directly with meta.slice.get()!

    Only works efficiently on large datasets with an inline executor!
    '''
    def __init__(self,
            reference,
            preferred_dtype=UDF.USE_NATIVE_DTYPE,
            validation_function=np.allclose):
        super().__init__(
            reference=reference,
            preferred_dtype=preferred_dtype,
            validation_function=validation_function
        )

    def get_preferred_input_dtype(self):
        return self.params.preferred_dtype

    def get_result_buffers(self):
        return {
            'seen': self.buffer(kind="nav", dtype=object),
        }

    def preprocess(self):
        self.results.seen[:] = [[] for _ in range(self.results.seen.size)]

    def process_tile(self, tile):
        """
        Verifies that the tile corresponds to the equivalent data
        in the reference array, and tracks which parts of the array
        have been 'seen'

        The elements of 'seen' are the slice tuples of each frame that
        have been provided to this method, potentially unordered.
        The remaining assumptions are that the flat nav dimension of
        the tile slice corresponds correctly to the view into the seen
        buffer, and that the frame slices are in fact correct for the
        data provided (which is partially ensured by 'validation_function'
        but could still fail on all-zeros data, for example).
        """
        # Cut out the flat nav dimension
        tile_sig_origin = self.meta.slice.origin[1:]
        tile_sig_shape = self.meta.slice.shape[1:]
        # construct slices
        frame_slices = tuple(slice(o, o + s) for o, s in zip(tile_sig_origin, tile_sig_shape))
        for i in range(self.results.seen.size):
            self.results.seen[i].append(frame_slices)
        assert self.params.validation_function(
            self.meta.slice.get(self.params.reference), tile
        )

    def merge(self, dest, src):
        for i in range(dest.seen.size):
            dest.seen[i].extend(src.seen[i])

    def _do_get_results(self):
        """
        Unpacks the slices seen for each frame and applies
        them to a frame-size bitmask to verify a complete
        tiling of each frame was recieved. If an ROI is provided
        then the roi==False frames should have None in place
        of their slices list, and are therefore checked against the ROI
        """
        results = super()._do_get_results()
        roi = self.meta.roi
        if roi is not None:
            roi = roi.reshape((-1,))
        sig_shape = self.meta.dataset_shape.sig.to_tuple()
        frame_mask = np.zeros(sig_shape, dtype=bool)
        for flat_idx, slices in enumerate(results['seen'].data.ravel()):
            frame_mask[:] = False
            if slices is None:
                assert roi is not None and not roi[flat_idx]
                continue
            # Check that ROI=true frames match
            if roi is not None:
                assert roi[flat_idx]
            # This could be optimized with some slice combination
            # but using a bitmask is completely robust and not
            # actually that slow to run
            for sl in slices:
                frame_mask[sl] = True
            assert frame_mask.all()
        assert (flat_idx + 1) == self.meta.dataset_shape.nav.size
        return results


def dataset_correction_verification(ds, roi, lt_ctx, exclude=None):
    """
    compare correct function w/ corrected pick
    """
    for i in range(1):
        shape = (-1, *tuple(ds.shape.sig))
        uncorr = CorrectionSet()
        data = lt_ctx.run_udf(udf=PickUDF(), dataset=ds, roi=roi, corrections=uncorr)

        gain = np.random.random(ds.shape.sig) + 1
        dark = np.random.random(ds.shape.sig) - 0.5

        if exclude is None:
            exclude = [
                (np.random.randint(0, s), np.random.randint(0, s))
                for s in tuple(ds.shape.sig)
            ]

        exclude_coo = sparse.COO(coords=exclude, data=True, shape=ds.shape.sig)
        corrset = CorrectionSet(dark=dark, gain=gain, excluded_pixels=exclude_coo)

        # This one uses native input data
        pick_res = lt_ctx.run_udf(udf=PickUDF(), dataset=ds, corrections=corrset, roi=roi)
        corrected = correct(
            buffer=data['intensity'].raw_data.reshape(shape),
            dark_image=dark,
            gain_map=gain,
            excluded_pixels=exclude,
            inplace=False
        )

        print("Exclude: ", exclude)

        print(pick_res['intensity'].raw_data.dtype)
        print(corrected.dtype)

        assert np.allclose(
            pick_res['intensity'].raw_data.reshape(shape),
            corrected
        )


def dataset_correction_masks(ds, roi, lt_ctx, exclude=None):
    """
    compare correction via sparse mask multiplication w/ correct function
    """
    for i in range(1):
        shape = (-1, *tuple(ds.shape.sig))
        uncorr = CorrectionSet()
        data = lt_ctx.run_udf(udf=PickUDF(), dataset=ds, roi=roi, corrections=uncorr)

        gain = np.random.random(ds.shape.sig) + 1
        dark = np.random.random(ds.shape.sig) - 0.5

        if exclude is None:
            exclude = [
                (np.random.randint(0, s), np.random.randint(0, s))
                for s in tuple(ds.shape.sig)
            ]

        exclude_coo = sparse.COO(coords=exclude, data=True, shape=ds.shape.sig)
        corrset = CorrectionSet(dark=dark, gain=gain, excluded_pixels=exclude_coo)

        def mask_factory():
            s = tuple(ds.shape.sig)
            return sparse.eye(np.prod(s)).reshape((-1, *s))

        # This one casts to float
        mask_res = lt_ctx.run_udf(
            udf=ApplyMasksUDF(mask_factory),
            dataset=ds,
            corrections=corrset,
            roi=roi,
        )
        # This one uses native input data
        corrected = correct(
            buffer=data['intensity'].raw_data.reshape(shape),
            dark_image=dark,
            gain_map=gain,
            excluded_pixels=exclude,
            inplace=False
        )

        print("Exclude: ", exclude)

        print(mask_res['intensity'].raw_data.dtype)
        print(corrected.dtype)

        assert np.allclose(
            mask_res['intensity'].raw_data.reshape(shape),
            corrected
        )


def get_testdata_path():
    return os.environ.get(
        'TESTDATA_BASE_PATH',
        os.path.normpath(
            os.path.join(os.path.dirname(__file__), '..', 'data')
        )
    )


class FakeBackend(IOBackend, id_="fake"):
    def get_impl(self):
        return FakeBackendImpl()


class FakeBackendImpl(IOBackendImpl):
    def get_tiles(
        self, tiling_scheme, fileset, read_ranges, roi, native_dtype, read_dtype, decoder,
        sync_offset, corrections, array_backend: ArrayBackend
    ):
        raise RuntimeError("nothing to see here")
        # to make this a generator, there needs to be a yield statement in
        # the body of the function, even if it is never executed:
        yield


def roi_as_sparse(roi):
    if roi is None:
        return roi
    return sparse.COO.from_numpy(roi, fill_value=False)


@contextmanager
def set_device_class(device_class):
    '''
    This context manager is designed to work with the inline executor.
    It simplifies running tests with several device classes by skipping
    unavailable device classes and handling setting and re-setting the environment variables
    correctly.
    '''
    prev_cuda_id = get_use_cuda()
    prev_cpu_id = get_use_cpu()
    try:
        if device_class in ('cupy', 'cuda'):
            d = detect()
            cudas = d['cudas']
            if not d['cudas']:
                pytest.skip(f"No CUDA device, skipping test with device class {device_class}.")
            if device_class == 'cupy' and not d['has_cupy']:
                pytest.skip(f"No CuPy, skipping test with device class {device_class}.")
            set_use_cuda(cudas[0])
        else:
            set_use_cpu(0)
        print(f'running with {device_class}')
        yield
    finally:
        if prev_cpu_id is not None:
            assert prev_cuda_id is None
            set_use_cpu(prev_cpu_id)
        elif prev_cuda_id is not None:
            assert prev_cpu_id is None
            set_use_cuda(prev_cuda_id)
        else:
            raise RuntimeError('No previous device ID, this should not happen.')



import os
import contextlib


@contextlib.contextmanager
def open_direct(path):
    """
    open `path` for reading with O_DIRECT
    """
    fh = os.open(path, os.O_RDONLY | os.O_DIRECT)
    f = open(fh, "rb", buffering=0)
    yield f
    f.close()


def readinto_direct(f, out):
    """
    read from `f` into numpy array `out`, returning a view
    for the read data
    """
    ret = f.readinto(out)
    if ret is None:
        raise OSError("could not readinto()")
    return out[:ret]



import os
import sys
import random

import numpy as np
import pytest

from libertem.executor.inline import InlineJobExecutor
from libertem.io.dataset.seq import (SEQDataSet, _load_xml_from_string, xml_defect_data_extractor,
                                     xml_map_sizes, xml_unbinned_map_maker, array_cropping,
                                     xml_defect_coord_extractor, xml_map_index_selector,
                                     xml_binned_map_maker)
from libertem.common import Shape
from libertem.common.buffers import reshaped_view
from libertem.udf.sumsigudf import SumSigUDF
from libertem.udf.raw import PickUDF
from libertem.io.dataset.base import TilingScheme, BufferedBackend, MMapBackend, DirectBackend
from libertem.io.corrections import CorrectionSet

from utils import get_testdata_path, ValidationUDF, roi_as_sparse
import defusedxml.ElementTree as ET

try:
    import pims
except ModuleNotFoundError:
    pims = None

SEQ_TESTDATA_PATH = os.path.join(get_testdata_path(), 'default.seq')
HAVE_SEQ_TESTDATA = os.path.exists(SEQ_TESTDATA_PATH)

needsdata = pytest.mark.skipif(not HAVE_SEQ_TESTDATA, reason="need .seq testdata")


@pytest.fixture
def default_seq(lt_ctx):
    nav_shape = (8, 8)
    ds = lt_ctx.load(
        "seq",
        path=SEQ_TESTDATA_PATH,
        nav_shape=nav_shape,
        io_backend=MMapBackend(),
    )

    ds.set_num_cores(4)
    assert tuple(ds.shape) == (8, 8, 128, 128)
    return ds


@pytest.fixture
def buffered_seq(lt_ctx):
    nav_shape = (8, 8)

    ds = lt_ctx.load(
        "seq",
        path=SEQ_TESTDATA_PATH,
        nav_shape=nav_shape,
        io_backend=BufferedBackend(),
    )

    ds.set_num_cores(4)
    return ds


@pytest.fixture
def direct_seq(lt_ctx):
    nav_shape = (8, 8)

    ds = lt_ctx.load(
        "seq",
        path=SEQ_TESTDATA_PATH,
        nav_shape=nav_shape,
        io_backend=DirectBackend(),
    )

    ds.set_num_cores(4)
    return ds


@pytest.fixture(scope='module')
def default_seq_raw():
    return np.array(pims.open(str(SEQ_TESTDATA_PATH))).reshape((8, 8, 128, 128))


@pytest.mark.skipif(pims is None, reason="No PIMS found")
@needsdata
def test_comparison(default_seq, default_seq_raw, lt_ctx_fast):
    corrset = CorrectionSet()
    udf = ValidationUDF(
        reference=reshaped_view(default_seq_raw, (-1, *tuple(default_seq.shape.sig)))
    )
    lt_ctx_fast.run_udf(udf=udf, dataset=default_seq, corrections=corrset)


@pytest.mark.skipif(pims is None, reason="No PIMS found")
@needsdata
def test_comparison_roi(default_seq, default_seq_raw, lt_ctx_fast):
    corrset = CorrectionSet()
    roi = np.random.choice(
        [True, False],
        size=tuple(default_seq.shape.nav),
        p=[0.5, 0.5]
    )
    udf = ValidationUDF(reference=default_seq_raw[roi])
    lt_ctx_fast.run_udf(udf=udf, dataset=default_seq, roi=roi, corrections=corrset)


@needsdata
def test_positive_sync_offset(default_seq, lt_ctx):
    udf = SumSigUDF()
    sync_offset = 2

    ds_with_offset = SEQDataSet(
        path=SEQ_TESTDATA_PATH, nav_shape=(8, 8), sync_offset=sync_offset
    )
    ds_with_offset.set_num_cores(4)
    ds_with_offset = ds_with_offset.initialize(lt_ctx.executor)
    ds_with_offset.check_valid()

    p0 = next(ds_with_offset.get_partitions())
    assert p0._start_frame == 2
    assert p0.slice.origin == (0, 0, 0)

    tileshape = Shape(
        (4,) + tuple(ds_with_offset.shape.sig),
        sig_dims=ds_with_offset.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=ds_with_offset.shape,
    )

    tiles = p0.get_tiles(tiling_scheme)
    t0 = next(tiles)
    assert tuple(t0.tile_slice.origin) == (0, 0, 0)
    for _ in tiles:
        pass

    for p in ds_with_offset.get_partitions():
        for t in p.get_tiles(tiling_scheme=tiling_scheme):
            pass

    assert p.slice.origin == (48, 0, 0)
    assert p.slice.shape[0] == 16

    result = lt_ctx.run_udf(dataset=default_seq, udf=udf)
    result = result['intensity'].raw_data[sync_offset:]

    result_with_offset = lt_ctx.run_udf(dataset=ds_with_offset, udf=udf)
    result_with_offset = result_with_offset['intensity'].raw_data[
                         :ds_with_offset._meta.image_count - sync_offset
                         ]

    assert np.allclose(result, result_with_offset)


def test_array_cropping():
    start_size = (1024, 1024)
    crop_to_this = (512, 512)
    offset = (600, 600)
    array = np.zeros(start_size)
    n_array = array_cropping(array, start_size, crop_to_this, offset)
    assert np.array_equal(n_array, array)


def test_xml_excluded_pixels_unbinned():
    xml_string = '''<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
                    <Configuration>
                        <PixelSize></PixelSize><DiffPixelSize></DiffPixelSize>
                        <BadPixels>
                        <BadPixelMap Rows="4096" Columns="4096">
                            <Defect Rows="2311-2312"/>
                            <Defect Rows="3413-3414"/>
                            <Defect Column="2311"/>
                        </BadPixelMap>
                        <BadPixelMap Binning="2" Rows="2048" Columns="2048">
                            <Defect Rows="1155-1156"/>
                            <Defect Rows="1706-1707"/>
                        </BadPixelMap>
                        </BadPixels>
                    </Configuration>
            '''
    metadata = {
        "UnbinnedFrameSizeX": 1024,
        "UnbinnedFrameSizeY": 1024,
        "OffsetX": 1536,
        "OffsetY": 1536,
        "HardwareBinning": 1
    }
    test_arr = np.zeros((1024, 1024), dtype=bool)
    test_arr[775] = True
    test_arr[:, 775] = True
    test_arr[776] = True
    expected_res = _load_xml_from_string(xml=xml_string, metadata=metadata)
    assert np.array_equal(expected_res.todense(), test_arr)


def test_xml_excluded_pixels_only_binned():
    xml_string = '''<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
                    <Configuration>
                        <PixelSize></PixelSize><DiffPixelSize></DiffPixelSize>
                        <BadPixels>
                            <BadPixelMap Rows="4096" Columns="4096">
                                <Defect Rows="2311-2312"/>
                                <Defect Row="600"/>
                                <Defect Columns="1310-1312"/>
                                <Defect Column="1300"/>
                                <Defect Row="100" Column="150"/>
                            </BadPixelMap>
                            <BadPixelMap Binning="2" Rows="2048" Columns="2048">
                                <Defect Rows="1155-1156"/>
                                <Defect Row="300"/>
                                <Defect Columns="655-656"/>
                                <Defect Column="650"/>
                                <Defect Row="50" Column="75"/>
                            </BadPixelMap>
                        </BadPixels>
                    </Configuration>
        '''
    metadata = {
        "UnbinnedFrameSizeX": 4096,
        "UnbinnedFrameSizeY": 4096,
        "OffsetX": 0,
        "OffsetY": 0,
        "HardwareBinning": 2
    }
    test_arr = np.zeros((2048, 2048), dtype=bool)
    test_arr[1155] = True
    test_arr[1156] = True
    test_arr[300] = True
    test_arr[50, 75] = True
    test_arr[:, 650] = True
    test_arr[:, 655:657] = True
    expected_res = _load_xml_from_string(xml=xml_string, metadata=metadata)
    assert np.array_equal(expected_res.todense(), test_arr)


def test_xml_excluded_pixels_binned_cropped():
    xml_string = '''<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
                    <Configuration>
                        <PixelSize></PixelSize><DiffPixelSize></DiffPixelSize>
                        <BadPixels>
                            <BadPixelMap Rows="4096" Columns="4096">
                                <Defect Rows="2311-2312"/>
                                <Defect Rows="3413-3414"/>
                                <Defect Columns="1310-1312"/>
                                <Defect Column="1300"/>
                                <Defect Row="100" Column="150"/>
                            </BadPixelMap>
                            <BadPixelMap Binning="2" Rows="2048" Columns="2048">
                                <Defect Rows="1250-1252"/>
                                <Defect Row="800"/>
                                <Defect Columns="768-770"/>
                                <Defect Column="1000"/>
                                <Defect Row="1200" Column="1100"/>
                            </BadPixelMap>
                        </BadPixels>
                    </Configuration>
        '''
    metadata = {
        "UnbinnedFrameSizeX": 1024,
        "UnbinnedFrameSizeY": 1024,
        "OffsetX": 1536,
        "OffsetY": 1536,
        "HardwareBinning": 2
    }
    test_arr = np.zeros((512, 512), dtype=bool)
    test_arr[482:485] = True
    test_arr[32] = True
    test_arr[:, 0:3] = True
    test_arr[:, 232] = True
    test_arr[432, 332] = True
    expected_res = _load_xml_from_string(xml=xml_string, metadata=metadata)
    assert np.array_equal(expected_res.todense(), test_arr)


def test_correct_bad_pixel_map_selector():
    xml_string = '''<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
                    <Configuration>
                        <PixelSize></PixelSize><DiffPixelSize></DiffPixelSize>
                        <BadPixels>
                            <BadPixelMap Rows="4096" Columns="4096">
                                <Defect Columns="1310-1312"/>
                            </BadPixelMap>
                            <BadPixelMap Rows="2048" Columns="2048">
                                <Defect Rows="1155-1156"/>
                                <Defect Rows="1706-1707"/>
                            </BadPixelMap>
                        </BadPixels>
                    </Configuration>
        '''
    metadata = {
        "UnbinnedFrameSizeX": 1024,
        "UnbinnedFrameSizeY": 1024,
        "OffsetX": 1536,
        "OffsetY": 1536,
        "HardwareBinning": 1
    }
    tree = ET.fromstring(xml_string)
    excluded_rows_dict = xml_defect_data_extractor(tree, metadata)
    assert excluded_rows_dict["cols"] == [['1310', '1312']]


def test_correct_bad_pixel_map_selector_2():
    xml_string = '''<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
                    <Configuration>
                        <PixelSize></PixelSize><DiffPixelSize></DiffPixelSize>
                        <BadPixels>
                            <BadPixelMap Rows="4096" Columns="4096">
                                <Defect Columns="1310-1312"/>
                            </BadPixelMap>
                            <BadPixelMap Binning="2" Rows="2048" Columns="2048">
                                <Defect Rows="1155-1156"/>
                            </BadPixelMap>
                        </BadPixels>
                    </Configuration>
        '''
    metadata = {
        "UnbinnedFrameSizeX": 1024,
        "UnbinnedFrameSizeY": 1024,
        "OffsetX": 1536,
        "OffsetY": 1536,
        "HardwareBinning": 2
    }
    tree = ET.fromstring(xml_string)
    excluded_rows_dict = xml_defect_data_extractor(tree, metadata)
    assert excluded_rows_dict["rows"] == [['1155', '1156']]


def test_map_size():
    xml = '''<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
          <BadPixels>
              <BadPixelMap Rows="4096" Columns="4096"></BadPixelMap>
              <BadPixelMap Binning="2" Rows="4080" Columns="4096"></BadPixelMap>
          </BadPixels>
          '''
    tree = ET.fromstring(xml)
    bad_pixel_maps = tree.findall('.//BadPixelMap')
    xy_map_sizes, map_sizes = xml_map_sizes(bad_pixel_maps)
    xy_map_sizes_expected = [(4096, 4096), (4096, 4080), (1, 2)]
    assert xy_map_sizes == xy_map_sizes_expected


def test_unbinned_map_maker():
    xml = '''<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
          <BadPixels>
              <BadPixelMap Rows="4096" Columns="4096"></BadPixelMap>
              <BadPixelMap Rows="2048" Columns="2048"></BadPixelMap>
              <BadPixelMap Binning="2" Rows="4080" Columns="4096"></BadPixelMap>
          </BadPixels>
         '''
    tree = ET.fromstring(xml)
    bad_pixel_maps = tree.findall('.//BadPixelMap')
    xy_map_sizes, _ = xml_map_sizes(bad_pixel_maps)
    unbinned_x, unbinned_y = xml_unbinned_map_maker(xy_map_sizes)
    assert unbinned_x == [4096, 2048, 0]
    assert unbinned_y == [4096, 2048, 0]


def test_binned_map_maker():
    xml = '''<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
          <BadPixels>
              <BadPixelMap Rows="4096" Columns="4096"></BadPixelMap>
              <BadPixelMap Rows="2048" Columns="2048"></BadPixelMap>
              <BadPixelMap Binning="2" Rows="4096" Columns="4096"></BadPixelMap>
          </BadPixels>
         '''
    tree = ET.fromstring(xml)
    bad_pixel_maps = tree.findall('.//BadPixelMap')
    xy_map_sizes, _ = xml_map_sizes(bad_pixel_maps)
    unbinned_x, unbinned_y = xml_binned_map_maker(xy_map_sizes)
    assert unbinned_x == [0, 0, 4096]
    assert unbinned_y == [0, 0, 4096]


def test_map_index_selector_case1():
    xml = '''<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
          <BadPixels>
              <BadPixelMap Rows="4096" Columns="4096"></BadPixelMap>
              <BadPixelMap Rows="2048" Columns="2048"></BadPixelMap>
              <BadPixelMap Binning="2" Rows="512" Columns="512"></BadPixelMap>
          </BadPixels>
         '''
    tree = ET.fromstring(xml)
    bad_pixel_maps = tree.findall('.//BadPixelMap')
    xy_map_sizes, map_sizes = xml_map_sizes(bad_pixel_maps)
    used_x, used_y = xml_unbinned_map_maker(xy_map_sizes)
    map_index = xml_map_index_selector(used_y)
    expected_index = 0
    assert map_index == expected_index


def test_map_index_selector_case2():
    xml = '''<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
          <BadPixels>
              <BadPixelMap Rows="4096" Columns="4096"></BadPixelMap>
              <BadPixelMap Rows="2048" Columns="2048"></BadPixelMap>
              <BadPixelMap Binning="2" Rows="512" Columns="512"></BadPixelMap>
          </BadPixels>
         '''
    tree = ET.fromstring(xml)
    bad_pixel_maps = tree.findall('.//BadPixelMap')
    xy_map_sizes, map_sizes = xml_map_sizes(bad_pixel_maps)
    used_x, used_y = xml_binned_map_maker(xy_map_sizes)
    map_index = xml_map_index_selector(used_y)
    print(used_y)
    expected_index = 2

    assert map_index == expected_index


def test_defect_extractor():
    xml = '''<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
          <BadPixels>
              <BadPixelMap Rows="4096" Columns="4096">
                  <Defect Rows="2311-2312" />
                  <Defect Row="1300" />
                  <Defect Columns="2300-2301" />
                  <Defect Column="600" />
                  <Defect Row="230" Column="100" />
              </BadPixelMap>
              <BadPixelMap Rows="2048" Columns="2048">
                    <Defect Column="10"/>
              </BadPixelMap>
              <BadPixelMap Binning="2" Rows="4080" Columns="4096"></BadPixelMap>
          </BadPixels>
          '''
    tree = ET.fromstring(xml)
    bad_pixel_maps = tree.findall('.//BadPixelMap')
    xy_map_sizes, map_sizes = xml_map_sizes(bad_pixel_maps)
    used_x, used_y = xml_unbinned_map_maker(xy_map_sizes)
    map_index = xml_map_index_selector(used_y)
    defects = xml_defect_coord_extractor(bad_pixel_maps[map_index], map_index, map_sizes)
    expected_defects = {"rows": [['2311', '2312'], ['1300']],
                        "cols": [['2300', '2301'], ['600']],
                        "pixels": [['100', '230']],
                        "size": (4096, 4096)
                        }
    assert defects == expected_defects


@needsdata
def test_negative_sync_offset(default_seq, lt_ctx):
    udf = SumSigUDF()
    sync_offset = -2

    ds_with_offset = SEQDataSet(
        path=SEQ_TESTDATA_PATH, nav_shape=(8, 8), sync_offset=sync_offset
    )
    ds_with_offset.set_num_cores(4)
    ds_with_offset = ds_with_offset.initialize(lt_ctx.executor)
    ds_with_offset.check_valid()

    p0 = next(ds_with_offset.get_partitions())
    assert p0._start_frame == -2
    assert p0.slice.origin == (0, 0, 0)

    tileshape = Shape(
        (4,) + tuple(ds_with_offset.shape.sig),
        sig_dims=ds_with_offset.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=ds_with_offset.shape,
    )

    tiles = p0.get_tiles(tiling_scheme)
    t0 = next(tiles)
    assert tuple(t0.tile_slice.origin) == (2, 0, 0)
    for _ in tiles:
        pass

    for p in ds_with_offset.get_partitions():
        for t in p.get_tiles(tiling_scheme=tiling_scheme):
            pass

    assert p.slice.origin == (48, 0, 0)
    assert p.slice.shape[0] == 16

    result = lt_ctx.run_udf(dataset=default_seq, udf=udf)
    result = result['intensity'].raw_data[:default_seq._meta.image_count - abs(sync_offset)]

    result_with_offset = lt_ctx.run_udf(dataset=ds_with_offset, udf=udf)
    result_with_offset = result_with_offset['intensity'].raw_data[abs(sync_offset):]

    assert np.allclose(result, result_with_offset)


@needsdata
def test_missing_frames(lt_ctx):
    nav_shape = (16, 8)

    ds = SEQDataSet(path=SEQ_TESTDATA_PATH, nav_shape=nav_shape)
    ds.set_num_cores(4)
    ds = ds.initialize(lt_ctx.executor)
    ds.check_valid()

    tileshape = Shape(
        (4,) + tuple(ds.shape.sig),
        sig_dims=ds.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=ds.shape,
    )

    for p in ds.get_partitions():
        for t in p.get_tiles(tiling_scheme=tiling_scheme):
            pass

    assert p._start_frame == 96
    assert p._num_frames == 32
    assert p.slice.origin == (96, 0, 0)
    assert p.slice.shape[0] == 32
    assert t.tile_slice.origin == (60, 0, 0)
    assert t.tile_slice.shape[0] == 4


@needsdata
def test_missing_data_with_positive_sync_offset(lt_ctx):
    nav_shape = (16, 8)
    sync_offset = 8

    ds = SEQDataSet(
        path=SEQ_TESTDATA_PATH, nav_shape=nav_shape, sync_offset=sync_offset
    )
    ds.set_num_cores(4)
    ds = ds.initialize(lt_ctx.executor)

    tileshape = Shape(
        (4,) + tuple(ds.shape.sig),
        sig_dims=ds.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=ds.shape,
    )

    for p in ds.get_partitions():
        for t in p.get_tiles(tiling_scheme=tiling_scheme):
            pass

    assert p._start_frame == 104
    assert p._num_frames == 32
    assert t.tile_slice.origin == (52, 0, 0)
    assert t.tile_slice.shape[0] == 4


@needsdata
def test_missing_data_with_negative_sync_offset(lt_ctx):
    nav_shape = (16, 8)
    sync_offset = -8

    ds = SEQDataSet(
        path=SEQ_TESTDATA_PATH, nav_shape=nav_shape, sync_offset=sync_offset
    )
    ds.set_num_cores(4)
    ds = ds.initialize(lt_ctx.executor)

    tileshape = Shape(
        (4,) + tuple(ds.shape.sig),
        sig_dims=ds.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=ds.shape,
    )

    for p in ds.get_partitions():
        for t in p.get_tiles(tiling_scheme=tiling_scheme):
            pass

    assert p._start_frame == 88
    assert p._num_frames == 32
    assert t.tile_slice.origin == (68, 0, 0)
    assert t.tile_slice.shape[0] == 4


@needsdata
def test_too_many_frames(lt_ctx):
    nav_shape = (4, 8)

    ds = SEQDataSet(path=SEQ_TESTDATA_PATH, nav_shape=nav_shape)
    ds.set_num_cores(4)
    ds = ds.initialize(lt_ctx.executor)
    ds.check_valid()

    tileshape = Shape(
        (4,) + tuple(ds.shape.sig),
        sig_dims=ds.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=ds.shape,
    )

    for p in ds.get_partitions():
        for t in p.get_tiles(tiling_scheme=tiling_scheme):
            pass


@needsdata
def test_positive_sync_offset_with_roi(default_seq, lt_ctx):
    udf = SumSigUDF()
    result = lt_ctx.run_udf(dataset=default_seq, udf=udf)
    result = result['intensity'].raw_data

    nav_shape = (8, 8)
    sync_offset = 2

    ds_with_offset = lt_ctx.load(
        "seq", path=SEQ_TESTDATA_PATH, nav_shape=nav_shape, sync_offset=sync_offset
    )

    roi = np.random.choice([False], (8, 8))
    roi[0:1] = True

    result_with_offset = lt_ctx.run_udf(dataset=ds_with_offset, udf=udf, roi=roi)
    result_with_offset = result_with_offset['intensity'].raw_data

    assert np.allclose(result[sync_offset:8 + sync_offset], result_with_offset)


@needsdata
def test_negative_sync_offset_with_roi(default_seq, lt_ctx):
    udf = SumSigUDF()
    result = lt_ctx.run_udf(dataset=default_seq, udf=udf)
    result = result['intensity'].raw_data

    nav_shape = (8, 8)
    sync_offset = -2

    ds_with_offset = lt_ctx.load(
        "seq", path=SEQ_TESTDATA_PATH, nav_shape=nav_shape, sync_offset=sync_offset
    )

    roi = np.random.choice([False], (8, 8))
    roi[0:1] = True

    result_with_offset = lt_ctx.run_udf(dataset=ds_with_offset, udf=udf, roi=roi)
    result_with_offset = result_with_offset['intensity'].raw_data

    assert np.allclose(result[:8 + sync_offset], result_with_offset[abs(sync_offset):])


@needsdata
def test_offset_smaller_than_image_count(lt_ctx):
    nav_shape = (8, 8)
    sync_offset = -65

    with pytest.raises(Exception) as e:
        lt_ctx.load(
            "seq",
            path=SEQ_TESTDATA_PATH,
            nav_shape=nav_shape,
            sync_offset=sync_offset
        )
    assert e.match(
        r"offset should be in \(-64, 64\), which is \(-image_count, image_count\)"
    )


@needsdata
def test_offset_greater_than_image_count(lt_ctx):
    nav_shape = (8, 8)
    sync_offset = 65

    with pytest.raises(Exception) as e:
        lt_ctx.load(
            "seq",
            path=SEQ_TESTDATA_PATH,
            nav_shape=nav_shape,
            sync_offset=sync_offset
        )
    assert e.match(
        r"offset should be in \(-64, 64\), which is \(-image_count, image_count\)"
    )


@needsdata
def test_reshape_nav(lt_ctx, default_seq):
    udf = SumSigUDF()

    ds_with_1d_nav = lt_ctx.load("seq", path=SEQ_TESTDATA_PATH, nav_shape=(64,))
    result_with_1d_nav = lt_ctx.run_udf(dataset=ds_with_1d_nav, udf=udf)
    result_with_1d_nav = result_with_1d_nav['intensity'].raw_data

    result_with_2d_nav = lt_ctx.run_udf(dataset=default_seq, udf=udf)
    result_with_2d_nav = result_with_2d_nav['intensity'].raw_data

    ds_with_3d_nav = lt_ctx.load("seq", path=SEQ_TESTDATA_PATH, nav_shape=(2, 4, 8))
    result_with_3d_nav = lt_ctx.run_udf(dataset=ds_with_3d_nav, udf=udf)
    result_with_3d_nav = result_with_3d_nav['intensity'].raw_data

    assert np.allclose(result_with_1d_nav, result_with_2d_nav, result_with_3d_nav)


@needsdata
def test_reshape_different_shapes(lt_ctx, default_seq):
    udf = SumSigUDF()

    result = lt_ctx.run_udf(dataset=default_seq, udf=udf)
    result = result['intensity'].raw_data

    ds_1 = lt_ctx.load("seq", path=SEQ_TESTDATA_PATH, nav_shape=(3, 6))
    result_1 = lt_ctx.run_udf(dataset=ds_1, udf=udf)
    result_1 = result_1['intensity'].raw_data

    assert np.allclose(result_1, result[:3 * 6])


@needsdata
def test_incorrect_sig_shape(lt_ctx):
    nav_shape = (8, 8)
    sig_shape = (5, 5)

    with pytest.raises(Exception) as e:
        lt_ctx.load(
            "seq",
            path=SEQ_TESTDATA_PATH,
            nav_shape=nav_shape,
            sig_shape=sig_shape
        )
    assert e.match(
        r"sig_shape must be of size: 16384"
    )


@needsdata
def test_scan_size_deprecation(lt_ctx):
    scan_size = (5, 5)

    with pytest.warns(FutureWarning):
        ds = lt_ctx.load(
            "seq",
            path=SEQ_TESTDATA_PATH,
            scan_size=scan_size,
        )
    assert tuple(ds.shape) == (5, 5, 128, 128)


@needsdata
def test_detect_non_seq(raw_with_zeros, lt_ctx):
    path = raw_with_zeros._path
    # raw_with_zeros is not a SEQ file, caused UnicodeDecodeError before:
    assert SEQDataSet.detect_params(path, InlineJobExecutor()) is False


@needsdata
def test_detect_seq(lt_ctx):
    path = SEQ_TESTDATA_PATH
    assert SEQDataSet.detect_params(path, lt_ctx.executor) is not False


# from utils import dataset_correction_verification
# FIXME test with actual test file

@needsdata
def test_compare_backends(lt_ctx, default_seq, buffered_seq):
    y = random.choice(range(default_seq.shape.nav[0]))
    x = random.choice(range(default_seq.shape.nav[1]))
    mm_f0 = lt_ctx.run(lt_ctx.create_pick_analysis(
        dataset=default_seq,
        x=x, y=y,
    )).intensity
    buffered_f0 = lt_ctx.run(lt_ctx.create_pick_analysis(
        dataset=buffered_seq,
        x=x, y=y,
    )).intensity

    assert np.allclose(mm_f0, buffered_f0)


@needsdata
@pytest.mark.skipif(
    sys.platform.startswith("darwin"),
    reason="No support for direct I/O on Mac OS X"
)
def test_compare_direct_to_mmap(lt_ctx, default_seq, direct_seq):
    y = random.choice(range(default_seq.shape.nav[0]))
    x = random.choice(range(default_seq.shape.nav[1]))
    mm_f0 = lt_ctx.run(lt_ctx.create_pick_analysis(
        dataset=default_seq,
        x=x, y=y,
    )).intensity
    buffered_f0 = lt_ctx.run(lt_ctx.create_pick_analysis(
        dataset=direct_seq,
        x=x, y=y,
    )).intensity

    assert np.allclose(mm_f0, buffered_f0)


@needsdata
@pytest.mark.parametrize(
    "as_sparse", (
        False,
        True
    ),
)
def test_compare_backends_sparse(lt_ctx, default_seq, buffered_seq, as_sparse):
    roi = np.zeros(default_seq.shape.nav, dtype=bool).reshape((-1,))
    roi[0] = True
    roi[1] = True
    roi[16] = True
    roi[32] = True
    roi[-1] = True
    if as_sparse:
        roi = roi_as_sparse(roi)
    mm_f0 = lt_ctx.run_udf(dataset=default_seq, udf=PickUDF(), roi=roi)['intensity']
    buffered_f0 = lt_ctx.run_udf(dataset=buffered_seq, udf=PickUDF(), roi=roi)['intensity']

    assert np.allclose(mm_f0, buffered_f0)


@needsdata
def test_scheme_too_large(default_seq):
    partitions = default_seq.get_partitions()
    p = next(partitions)
    depth = p.shape[0]

    # we make a tileshape that is too large for the partition here:
    tileshape = Shape(
        (depth + 1,) + tuple(default_seq.shape.sig),
        sig_dims=default_seq.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=default_seq.shape,
    )

    # tile shape is clamped to partition shape:
    tiles = p.get_tiles(tiling_scheme=tiling_scheme)
    t = next(tiles)
    assert tuple(t.tile_slice.shape) == tuple((depth,) + default_seq.shape.sig)
    for _ in tiles:
        pass


def test_bad_params(ds_params_tester, standard_bad_ds_params):
    args = ("seq", SEQ_TESTDATA_PATH)
    for params in standard_bad_ds_params:
        if "nav_shape" not in params:
            params['nav_shape'] = (8, 8)
        ds_params_tester(*args, **params)



import numpy as np
import pytest
import cloudpickle

from sparseconverter import BACKENDS
from libertem.udf.base import UDF, NoOpUDF, UDFPartRunner, UDFParams, UDFMeta
from libertem.common.executor import Environment
from libertem.common.udf import UDFMethod
from libertem.io.dataset.memory import MemoryDataSet
from libertem.io.dataset.base import TilingScheme, DataTile
from libertem.io.dataset.raw import RawFileDataSet, RawPartition
from libertem.utils.devices import detect
from libertem.common.exceptions import UDFException
from libertem.common.backend import set_use_cpu, set_use_cuda
from libertem.common.buffers import reshaped_view
from libertem.udf.sumsigudf import SumSigUDF
from libertem.udf.sum import SumUDF

from utils import _mk_random, ValidationUDF


def test_validation(lt_ctx):
    """
    Test that the ValidationUDF works as designed
    """
    data = _mk_random(size=(16, 16, 16, 16), dtype="float32")
    dataset = MemoryDataSet(data=data, tileshape=(1, 16, 16),
                            num_partitions=2, sig_dims=2)

    udf = ValidationUDF(reference=data.reshape((-1, 16, 16)))
    res = lt_ctx.run_udf(dataset=dataset, udf=udf)
    assert res['seen'].data.shape == (16, 16)

    with pytest.raises(AssertionError):
        data2 = data.copy()
        data2[7, 9, 13, 11] += 0.1
        udf = ValidationUDF(reference=data2.reshape((-1, 16, 16)))
        res = lt_ctx.run_udf(dataset=dataset, udf=udf)

    with pytest.raises(AssertionError):
        def badcompare(a, b):
            return False

        udf = ValidationUDF(
            reference=data.reshape((-1, 16, 16)),
            validation_function=badcompare
        )
        res = lt_ctx.run_udf(dataset=dataset, udf=udf)


class MissingFramesDataset(RawFileDataSet):
    def get_partitions(self):
        fileset = self._get_fileset()
        for part_slice, start, stop in self.get_slices():
            yield MissingFramesPartition(
                meta=self._meta,
                fileset=fileset,
                partition_slice=part_slice,
                start_frame=start,
                num_frames=stop - start,
                io_backend=self.get_io_backend(),
            )


class MissingFramesPartition(RawPartition):
    def get_tiles(self, *args, **kwargs):
        """
        Skips 50% of tiles, guaranteed to skip at least the first
        """
        if self._start_frame < self.meta.image_count:
            for tile_idx, tile in enumerate(super().get_tiles(*args, **kwargs)):
                if tile_idx % 2:
                    yield tile


@pytest.mark.parametrize(
    "with_roi", (True, False)
)
def test_validation_skips(default_raw: RawFileDataSet, default_raw_data: np.ndarray,
                          lt_ctx_fast, with_roi):
    ds = MissingFramesDataset(default_raw._path,
                              dtype=default_raw.meta.dtype,
                              nav_shape=default_raw.meta.shape.nav,
                              sig_shape=default_raw.meta.shape.sig)
    ds.initialize(lt_ctx_fast.executor)

    if with_roi:
        roi = np.zeros(ds.shape.nav, dtype=bool)
        roi[:1] = True
    else:
        roi = None

    udf = ValidationUDF(
        reference=reshaped_view(default_raw_data, (-1, *tuple(ds.shape.sig)))
    )
    with pytest.raises(AssertionError):
        lt_ctx_fast.run_udf(udf=udf, dataset=ds, roi=roi)


class PixelsumUDF(UDF):
    def get_result_buffers(self):
        return {
            'pixelsum': self.buffer(
                kind="nav", dtype="float32"
            )
        }

    def process_frame(self, frame):
        assert frame.shape == (16, 16)
        assert self.results.pixelsum.shape == (1,)
        self.results.pixelsum[:] = np.sum(frame)


def test_sum_frames(lt_ctx):
    """
    Test sum over the pixels for 2-dimensional dataset

    Parameters
    ----------
    lt_ctx
        Context class for loading dataset and creating jobs on them

    """
    data = _mk_random(size=(16, 16, 16, 16), dtype="float32")
    dataset = MemoryDataSet(data=data, tileshape=(1, 16, 16),
                            num_partitions=2, sig_dims=2)

    pixelsum = PixelsumUDF()
    res = lt_ctx.run_udf(dataset=dataset, udf=pixelsum)
    assert 'pixelsum' in res
    print(data.shape, res['pixelsum'].data.shape)
    assert np.allclose(res['pixelsum'].data, np.sum(data, axis=(2, 3)))


def test_3d_ds(lt_ctx):
    """
    Test sum over the pixels for 3-dimensional dataset

    Parameters
    ----------
    lt_ctx
        Context class for loading dataset and creating jobs on them
    """
    data = _mk_random(size=(16 * 16, 16, 16), dtype="float32")
    dataset = MemoryDataSet(data=data, tileshape=(1, 16, 16),
                            num_partitions=2, sig_dims=2)

    pixelsum = PixelsumUDF()
    res = lt_ctx.run_udf(dataset=dataset, udf=pixelsum)
    assert 'pixelsum' in res
    print(data.shape, res['pixelsum'].data.shape)
    assert np.allclose(res['pixelsum'].data, np.sum(data, axis=(1, 2)))


def test_kind_single(lt_ctx):
    """
    Test buffer type kind='single'

    Parameters
    ----------
    lt_ctx
        Context class for loading dataset and creating jobs on them
    """
    data = _mk_random(size=(16, 16, 16, 16), dtype="float32")
    dataset = MemoryDataSet(data=data, tileshape=(2, 16, 16),
                            num_partitions=2, sig_dims=2)

    class CounterUDF(UDF):
        def get_result_buffers(self):
            return {
                'counter': self.buffer(
                    kind="single", dtype="uint32"
                ),
                'sum_frame': self.buffer(
                    kind="single", extra_shape=(16,), dtype="float32"
                )
            }

        def process_frame(self, frame):
            self.results.counter[:] += 1
            self.results.sum_frame[:] += np.sum(frame, axis=1)

        def merge(self, dest, src):
            dest.counter[:] += src.counter
            dest.sum_frame[:] += src.sum_frame

    counter = CounterUDF()
    res = lt_ctx.run_udf(dataset=dataset, udf=counter)
    assert 'counter' in res
    assert 'sum_frame' in res
    assert res['counter'].data.shape == (1,)
    assert res['counter'].data == 16 * 16
    assert res['sum_frame'].data.shape == (16,)
    assert np.allclose(res['sum_frame'].data, np.sum(data, axis=(0, 1, 3)))


def test_bad_merge_now_actually_good(lt_ctx):
    """
    Test previously bad example of updating buffer
    """
    data = _mk_random(size=(16 * 16, 16, 16), dtype="float32")
    dataset = MemoryDataSet(data=data, tileshape=(1, 16, 16),
                            num_partitions=2, sig_dims=2)

    class BadmergeUDF(UDF):
        def get_result_buffers(self):
            return {
                'pixelsum': self.buffer(
                    kind="nav", dtype="float32"
                )
            }

        def process_frame(self, frame):
            self.results.pixelsum[:] = np.sum(frame)

        def merge(self, dest, src):
            # this used not to work, now it should:
            dest.pixelsum = src.pixelsum

    bm = BadmergeUDF()
    assert np.allclose(
        lt_ctx.run_udf(dataset=dataset, udf=bm)['pixelsum'],
        lt_ctx.run_udf(dataset=dataset, udf=SumSigUDF())['intensity'],
    )


def test_plusequal_without_slice_assignment(lt_ctx, default_raw):
    class SigNoSlicePLusEqual(UDF):
        def get_result_buffers(self):
            return {
                'sum': self.buffer(
                    kind="sig", dtype="float32"
                )
            }

        def process_frame(self, frame):
            self.results.sum += frame

        def merge(self, dest, src):
            dest.sum += src.sum

    bm = SigNoSlicePLusEqual()
    assert np.allclose(
        lt_ctx.run_udf(dataset=default_raw, udf=bm)['sum'],
        lt_ctx.run_udf(dataset=default_raw, udf=SumUDF())['intensity'],
    )


def test_no_default_merge(lt_ctx):
    """
    Test forgotten merge function if not :code:`kind='nav'`.
    """
    data = _mk_random(size=(16 * 16, 16, 16), dtype="float32")
    dataset = MemoryDataSet(data=data, tileshape=(1, 16, 16),
                            num_partitions=2, sig_dims=2)

    class NodefaultUDF(UDF):
        def get_result_buffers(self):
            return {
                'pixelsum_nav': self.buffer(
                    kind="nav", dtype="float32"
                ),
                'pixelsum': self.buffer(
                    kind="sig", dtype="float32"
                )
            }

        def process_frame(self, frame):
            self.results.pixelsum[:] += frame

    with pytest.raises(NotImplementedError):
        nd = NodefaultUDF()
        lt_ctx.run_udf(dataset=dataset, udf=nd)


def test_extra_dimension_shape(lt_ctx):
    """
    Test sum over the pixels for 2-dimensional dataset

    Parameters
    ----------
    lt_ctx
        Context class for loading dataset and creating jobs on them

    """
    data = _mk_random(size=(16, 16, 16, 16), dtype="float32")
    dataset = MemoryDataSet(data=data, tileshape=(1, 16, 16),
                            num_partitions=2, sig_dims=2)

    class ExtraShapeUDF(UDF):
        def get_result_buffers(self):
            return {
                'test': self.buffer(
                    kind="nav", extra_shape=(2,), dtype="float32"
                ),
                'test2': self.buffer(
                    kind="sig", extra_shape=(2,), dtype="float32"
                ),
                'test3': self.buffer(
                    kind="single", extra_shape=(2,), dtype="float32"
                )
            }

        def process_frame(self, frame):
            self.results.test[:] = (1, 2)
            self.results.test2[:] += np.ones(tuple(self.meta.dataset_shape.sig) + (2, ))
            self.results.test3[:] += (1, 2)

        def merge(self, dest, src):
            dest.test[:] = src.test[:]
            dest.test2[:] += src.test2[:]
            dest.test3[:] += src.test3[:]

    extra = ExtraShapeUDF()
    res = lt_ctx.run_udf(dataset=dataset, udf=extra)

    navcount = np.prod(tuple(dataset.shape.nav))

    print(data.shape, res['test'].data.shape)
    assert res['test'].data.shape == tuple(dataset.shape.nav) + (2,)
    assert res['test2'].data.shape == tuple(dataset.shape.sig) + (2,)
    assert res['test3'].data.shape == (2,)
    assert res['test'].extra_shape == (2,)
    assert np.allclose(res['test'].data, (1, 2))
    assert np.allclose(res['test2'].data, navcount)
    assert np.allclose(res['test3'].data, (navcount, 2*navcount))


def test_roi_1(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype="float32")
    dataset = MemoryDataSet(data=data, tileshape=(3, 16, 16),
                            num_partitions=4, sig_dims=2)
    mask = np.random.choice([True, False], size=(16, 16))

    pixelsum = PixelsumUDF()
    res = lt_ctx.run_udf(dataset=dataset, udf=pixelsum, roi=mask)
    assert 'pixelsum' in res
    print(data.shape, res['pixelsum'].data.shape)
    expected = np.sum(data[mask, ...], axis=(-1, -2))
    assert np.allclose(res['pixelsum'].raw_data, expected)


def test_roi_all_zeros(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype="float32")
    dataset = MemoryDataSet(data=data, tileshape=(3, 16, 16),
                            num_partitions=16, sig_dims=2)
    mask = np.zeros(data.shape[:2], dtype=bool)

    pixelsum = PixelsumUDF()
    res = lt_ctx.run_udf(dataset=dataset, udf=pixelsum, roi=mask)
    assert 'pixelsum' in res
    print(data.shape, res['pixelsum'].data.shape)
    expected = np.sum(data[mask, ...], axis=(-1, -2))
    assert np.allclose(res['pixelsum'].raw_data, expected)


def test_roi_all_ones(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype="float32")
    dataset = MemoryDataSet(data=data, tileshape=(3, 16, 16),
                            num_partitions=16, sig_dims=2)
    mask = np.ones(data.shape[:2], dtype=bool)

    pixelsum = PixelsumUDF()
    res = lt_ctx.run_udf(dataset=dataset, udf=pixelsum, roi=mask)
    assert 'pixelsum' in res
    print(data.shape, res['pixelsum'].data.shape)
    expected = np.sum(data[mask, ...], axis=(-1, -2))
    assert np.allclose(res['pixelsum'].raw_data, expected)


def test_roi_some_zeros(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype="float32")
    dataset = MemoryDataSet(data=data, tileshape=(3, 16, 16),
                            num_partitions=16, sig_dims=2)
    mask = np.zeros(data.shape[:2], dtype=bool)
    mask[0] = True

    pixelsum = PixelsumUDF()
    res = lt_ctx.run_udf(dataset=dataset, udf=pixelsum, roi=mask)
    assert 'pixelsum' in res
    print(data.shape, res['pixelsum'].data.shape)
    expected = np.sum(data[mask, ...], axis=(-1, -2))
    assert np.allclose(res['pixelsum'].raw_data, expected)


def test_roi_extra_dimension_shape(lt_ctx):
    """
    Test sum over the pixels for 2-dimensional dataset

    Parameters
    ----------
    lt_ctx
        Context class for loading dataset and creating jobs on them

    """
    data = _mk_random(size=(16, 16, 16, 16), dtype="float32")
    dataset = MemoryDataSet(data=data, tileshape=(1, 16, 16),
                            num_partitions=2, sig_dims=2)

    class ExtraShapeUDF(UDF):
        def get_result_buffers(self):
            return {
                'test': self.buffer(
                    kind="nav", extra_shape=(2,), dtype="float32"
                ),
                'test2': self.buffer(
                    kind="sig", extra_shape=(2,), dtype="float32"
                ),
                'test3': self.buffer(
                    kind="single", extra_shape=(2,), dtype="float32"
                )
            }

        def process_frame(self, frame):
            self.results.test[:] = (1, 2)
            self.results.test2[:] += np.ones(tuple(self.meta.dataset_shape.sig) + (2, ))
            self.results.test3[:] += (1, 2)

        def merge(self, dest, src):
            dest.test[:] = src.test[:]
            dest.test2[:] += src.test2[:]
            dest.test3[:] += src.test3[:]

    extra = ExtraShapeUDF()
    roi = _mk_random(size=dataset.shape.nav, dtype=bool)
    res = lt_ctx.run_udf(dataset=dataset, udf=extra, roi=roi)

    navcount = np.count_nonzero(roi)
    print(navcount)

    print(data.shape, res['test'].data.shape)
    assert res['test'].data.shape == tuple(dataset.shape.nav) + (2,)
    assert res['test2'].data.shape == tuple(dataset.shape.sig) + (2,)
    assert res['test3'].data.shape == (2,)
    assert np.allclose(res['test'].raw_data, (1, 2))
    assert np.allclose(res['test2'].raw_data, navcount)
    assert np.allclose(res['test3'].raw_data, (navcount, 2*navcount))


def test_udf_pickle(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype="float32")
    dataset = MemoryDataSet(data=data, tileshape=(3, 16, 16),
                            num_partitions=16, sig_dims=2)

    partition = next(dataset.get_partitions())
    pixelsum = PixelsumUDF()
    meta = UDFMeta(
        partition_slice=partition.slice,
        dataset_shape=dataset.shape,
        roi=None,
        dataset_dtype="float32",
        input_dtype="float32",
        device_class="cpu",
    )
    pixelsum.set_backend("numpy")
    pixelsum.set_meta(meta)
    pixelsum.init_result_buffers()
    pixelsum.allocate_for_part(partition, None)
    cloudpickle.loads(cloudpickle.dumps(pixelsum))


class ExtraShapeWithZero(UDF):
    def get_result_buffers(self):
        return {
            'testnav': self.buffer(
                kind="nav", dtype="float32", extra_shape=(0,)
            ),
            'testsig': self.buffer(
                kind="sig", dtype="float32", extra_shape=(0,)
            ),
            'testsingle': self.buffer(
                kind="single", dtype="float32", extra_shape=(0,)
            )
        }

    def process_frame(self, frame):
        pass

    def merge(self, dest, src):
        pass


def test_extra_shape_with_zero(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype="float32")
    dataset = MemoryDataSet(data=data, tileshape=(1, 16, 16),
                            num_partitions=2, sig_dims=2)

    udf = ExtraShapeWithZero()
    res = lt_ctx.run_udf(dataset=dataset, udf=udf)

    assert res['testnav'].data.size == 0
    assert res['testnav'].data.shape == tuple(dataset.shape.nav) + (0,)
    assert res['testsig'].data.size == 0
    assert res['testsig'].data.shape == tuple(dataset.shape.sig) + (0,)
    assert res['testsingle'].data.size == 0
    assert res['testsingle'].data.shape == (0,)


@pytest.mark.parametrize(
    'preferred_dtype,data_dtype,expected_dtype',
    [
        (bool, np.uint16, None),
        (None, bool, np.float32),
        (None, np.complex64, np.complex64),
        (None, np.complex128, np.complex128),
        (np.int16, np.uint16, None),
        (UDF.USE_NATIVE_DTYPE, np.int32, np.int32)
    ]
)
def test_dtypes(lt_ctx, preferred_dtype, data_dtype, expected_dtype):
    class DebugDTypeUDF(UDF):
        def __init__(self, preferred_dtype):
            super().__init__(preferred_dtype=preferred_dtype)

        def get_preferred_input_dtype(self):
            if self.params.preferred_dtype is None:
                return super().get_preferred_input_dtype()
            else:
                return self.params.preferred_dtype

        def get_result_buffers(self):
            return {
                'dtype': self.buffer(
                    kind='nav', dtype='object'
                ),
                'input_dtype': self.buffer(
                    kind='single', dtype=self.meta.input_dtype
                ),
                'dataset_dtype': self.buffer(
                    kind='single', dtype=self.meta.dataset_dtype
                )
            }

        def process_frame(self, frame):
            self.results.dtype[:] = frame.dtype
            assert frame.dtype == self.meta.input_dtype
            assert self.results.input_dtype.dtype == self.meta.input_dtype
            assert self.results.dataset_dtype.dtype == self.meta.dataset_dtype

        def merge(self, dest, src):
            dest.dtype[:] = src.dtype[:]

    if expected_dtype is None:
        expected_dtype = np.result_type(preferred_dtype, data_dtype)
    data = np.zeros(shape=(1, 1), dtype=data_dtype)
    dataset = MemoryDataSet(data=data, sig_dims=1)

    udf = DebugDTypeUDF(preferred_dtype=preferred_dtype)

    res = lt_ctx.run_udf(udf=udf, dataset=dataset)

    print(res['dtype'].data[0])
    assert res['dtype'].data[0] == expected_dtype
    assert udf.meta.dataset_dtype == data_dtype
    assert udf.meta.input_dtype == expected_dtype
    assert res['input_dtype'].data.dtype == expected_dtype
    assert res['dataset_dtype'].data.dtype == data_dtype


def test_with_progress_bar(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype="float32")
    dataset = MemoryDataSet(data=data, tileshape=(1, 16, 16),
                            num_partitions=2, sig_dims=2)

    pixelsum = PixelsumUDF()
    lt_ctx.run_udf(dataset=dataset, udf=pixelsum, progress=True)
    # TODO: maybe assert that some output happened on stderr?


class ReshapedViewUDF(UDF):
    def get_result_buffers(self):
        return {
            "sigbuf": self.buffer(kind="sig", dtype=int, where="device")
        }

    def process_tile(self, tile):
        reshaped_view(tile, (tile.shape[0], -1))
        flat_buf = reshaped_view(self.results.sigbuf, (-1, ))
        flat_buf[:] = 1

    def merge(self, dest, src):
        dest.sigbuf[:] = src.sigbuf[:]

    def get_backends(self):
        return ('numpy', 'cupy')


@pytest.mark.parametrize(
    'backend', ['numpy', 'cupy']
)
def test_noncontiguous_tiles(lt_ctx, backend):
    if backend == 'cupy':
        d = detect()
        cudas = detect()['cudas']
        if not d['cudas'] or not d['has_cupy']:
            pytest.skip("No CUDA device or no CuPy, skipping CuPy test")

    data = _mk_random(size=(30, 3, 7), dtype="float32")
    dataset = MemoryDataSet(
        data=data, tileshape=(3, 2, 2),
        num_partitions=2, sig_dims=2
    )
    try:
        if backend == 'cupy':
            set_use_cuda(cudas[0])
        udf = ReshapedViewUDF()
        res = lt_ctx.run_udf(udf=udf, dataset=dataset)
        partition = next(dataset.get_partitions())
        p_udf = udf.copy_for_partition(partition=partition, roi=None)
        tiling_scheme = TilingScheme.make_for_shape(
            tileshape=dataset.tileshape,
            dataset_shape=dataset.shape,
        )
        # Enabling debug=True checks for disjoint cache keys
        params = UDFParams.from_udfs(
            udfs=[udf],
            roi=None,
            corrections=None,
            tiling_scheme=tiling_scheme,
        )
        UDFPartRunner([p_udf], debug=True).run_for_partition(
            partition=partition,
            params=params,
            env=Environment(threads_per_worker=1, threaded_executor=False),
            backend_choice=BACKENDS,
        )

    finally:
        set_use_cpu(0)

    assert np.all(res["sigbuf"].data == 1)


class CheckSigSlice(UDF):
    def get_result_buffers(self):
        return {
            "checksum": self.buffer(kind="nav", dtype=float, where="device")
        }

    def process_tile(self, tile):
        assert tile.shape[1:] == tuple(self.meta.sig_slice.shape)
        assert self.meta.slice.sig == self.meta.sig_slice
        assert self.meta.tiling_scheme[self.meta.tiling_scheme_idx] == self.meta.sig_slice
        if isinstance(tile, DataTile):
            # this is technically an internal interface, we test here
            # as a sanity check:
            assert tile.scheme_idx == self.meta.tiling_scheme_idx
        self.results.checksum[:] += tile.sum(axis=(-1, -2))

    def get_backends(self):
        return ('numpy', 'cupy')


class CheckSigSlicePartition(UDF):
    def get_result_buffers(self):
        return {
            "checksum": self.buffer(kind="nav", dtype=float, where="device")
        }

    def process_partition(self, tile):
        assert tile.shape[1:] == tuple(self.meta.sig_slice.shape)
        assert self.meta.slice.sig == self.meta.sig_slice
        assert self.meta.tiling_scheme[self.meta.tiling_scheme_idx] == self.meta.sig_slice
        assert self.meta.slice.shape == self.meta.partition_shape
        assert self.meta.slice == self.meta._partition_slice
        if isinstance(tile, DataTile):
            # this is technically an internal interface, we test here
            # as a sanity check:
            assert tile.scheme_idx == self.meta.tiling_scheme_idx
        self.results.checksum[:] += tile.sum(axis=(-1, -2))

    def get_backends(self):
        return ('numpy', 'cupy')


class CheckSigSliceFrame(UDF):
    def get_result_buffers(self):
        return {
            "checksum": self.buffer(kind="nav", dtype=float, where="device")
        }

    def process_frame(self, frame):
        print(frame.shape)
        assert frame.shape == tuple(self.meta.sig_slice.shape)
        assert self.meta.slice.sig == self.meta.sig_slice
        assert self.meta.slice.sig.shape == self.meta.dataset_shape.sig
        assert self.meta.tiling_scheme[self.meta.tiling_scheme_idx] == self.meta.sig_slice
        if isinstance(frame, DataTile):
            # this is technically an internal interface, we test here
            # as a sanity check:
            assert frame.scheme_idx == self.meta.tiling_scheme_idx
        self.results.checksum[:] += frame.sum(axis=(-1, -2))

    def get_backends(self):
        return ('numpy', 'cupy')


@pytest.mark.parametrize(
    'udf_class,tileshape,success', [
        (CheckSigSliceFrame, (3, 3, 5), False),
        (CheckSigSliceFrame, (3, 3, 7), True),
        (CheckSigSlicePartition, (3, 3, 7), True),
        (CheckSigSlicePartition, (15, 3, 7), True),
        (CheckSigSlice, (3, 2, 2), True),
    ],
)
@pytest.mark.parametrize(
    'backend', ['numpy', 'cupy'],
)
def test_sig_slice(lt_ctx, backend, udf_class, tileshape, success):
    if backend == 'cupy':
        d = detect()
        cudas = detect()['cudas']
        if not d['cudas'] or not d['has_cupy']:
            pytest.skip("No CUDA device or no CuPy, skipping CuPy test")

    data = _mk_random(size=(30, 3, 7), dtype="float32")
    ref_res = data.sum(axis=(-1, -2))
    dataset = MemoryDataSet(
        data=data, tileshape=tileshape,
        num_partitions=2, sig_dims=2
    )
    try:
        if backend == 'cupy':
            set_use_cuda(cudas[0])
        udf = udf_class()
        if success:
            res = lt_ctx.run_udf(udf=udf, dataset=dataset)
            assert np.allclose(res['checksum'].raw_data, ref_res)
        else:
            with pytest.raises(Exception):
                lt_ctx.run_udf(udf=udf, dataset=dataset)
    finally:
        set_use_cpu(0)


class ParamsCheckUDF(UDF):
    def get_result_buffers(self):
        return {
            "null": self.buffer(kind="nav", dtype=float)
        }

    def process_frame(self, frame):
        assert isinstance(self.params.int_param, int)
        assert self.params.get('is_222', default=None) == 222
        assert self.params.get('is_missing', default=333) == 333
        assert self.params.get('is_missing') is None


def test_params_check(lt_ctx):
    ds = lt_ctx.load('memory', data=np.ones((2, 2, 4, 4)))
    udf = ParamsCheckUDF(int_param=5, is_222=222)
    lt_ctx.run_udf(dataset=ds, udf=udf)


class TileAttrsUDF1(NoOpUDF):
    def process_tile(self, tile):
        tile.scheme_idx


class TileAttrsUDF2(NoOpUDF):
    def process_tile(self, tile):
        tile.tile_slice


@pytest.mark.parametrize(
    'cls', [TileAttrsUDF1, TileAttrsUDF2]
)
def test_previous_attrs(lt_ctx, cls):
    ds = lt_ctx.load('memory', data=np.ones((2, 2, 4, 4)))
    # Check that we get an attribute error that points to
    # the current alternative
    with pytest.raises(AttributeError, match='self.meta.'):
        lt_ctx.run_udf(dataset=ds, udf=cls())


class BadGetMethodUDF(UDF):
    def get_result_buffers(self):
        return {}

    def get_method(self):
        return self.params.get('method', 42)


@pytest.mark.parametrize(
    'method',
    [
        42,
        UDFMethod.FRAME,
        UDFMethod.PARTITION,
        UDFMethod.TILE,
    ]
)
def test_bad_custom_get_method(lt_ctx, method):
    # UDF implements no methods, so even if get_method()
    # returns a valid member of the enum the UDF interface
    # should raise the exception somewhere during initialization
    # also checks that the invalid return '42' causes the raise
    ds = lt_ctx.load('memory', data=np.ones((2, 2, 4, 4)))
    with pytest.raises(UDFException):
        lt_ctx.run_udf(dataset=ds, udf=BadGetMethodUDF(method=method))


class NoImplemUDF(UDF):
    ...


def test_no_implementation_default_raises():
    # Checks the default get_method() raises when no method is implemented
    with pytest.raises(TypeError):
        NoImplemUDF().get_method()


class MultiImplemUDF(UDF):
    def get_result_buffers(self):
        return {'val': self.buffer(kind='nav', dtype=int)}

    def process_tile(self, tile):
        self.results.val[:] = 42

    def process_frame(self, frame):
        self.results.val[:] = 53

    def process_partition(self, partition):
        self.results.val[:] = 64

    def get_method(self):
        if self.params.method == 'frame':
            return UDF.UDF_METHOD.FRAME
        elif self.params.method == 'partition':
            return UDF.UDF_METHOD.PARTITION
        else:
            return UDF.UDF_METHOD.TILE


def test_multi_implem(lt_ctx):
    ds = lt_ctx.load('memory', data=np.ones((2, 2, 4, 4)))
    method = None
    res = lt_ctx.run_udf(dataset=ds, udf=MultiImplemUDF(method=method))
    assert (res['val'].data == 42).all()
    method = 'frame'
    res = lt_ctx.run_udf(dataset=ds, udf=MultiImplemUDF(method=method))
    assert (res['val'].data == 53).all()
    method = 'partition'
    res = lt_ctx.run_udf(dataset=ds, udf=MultiImplemUDF(method=method))
    assert (res['val'].data == 64).all()



import os
import warnings

import numpy as np

from libertem.common.math import prod
from libertem.common import Shape
from .base import (
    DataSet, DataSetException, DataSetMeta,
    BasePartition, FileSet, File, IOBackend,
)
from libertem.common.messageconverter import MessageConverter

MAGIC_EXPECT = 258


class BLODatasetParams(MessageConverter):
    SCHEMA = {
      "$schema": "http://json-schema.org/draft-07/schema#",
      "$id": "http://libertem.org/BLODatasetParams.schema.json",
      "title": "BLODatasetParams",
      "type": "object",
      "properties": {
        "type": {"const": "BLO"},
        "path": {"type": "string"},
        "nav_shape": {
            "type": "array",
            "items": {"type": "number", "minimum": 1},
            "minItems": 2,
            "maxItems": 2
        },
        "sig_shape": {
            "type": "array",
            "items": {"type": "number", "minimum": 1},
            "minItems": 2,
            "maxItems": 2
        },
        "sync_offset": {"type": "number"},
        "io_backend": {
            "enum": IOBackend.get_supported(),
        },
      },
      "required": ["type", "path"],
    }

    def convert_to_python(self, raw_data):
        data = {
            k: raw_data[k]
            for k in ["path"]
        }
        if "nav_shape" in raw_data:
            data["nav_shape"] = tuple(raw_data["nav_shape"])
        if "sig_shape" in raw_data:
            data["sig_shape"] = tuple(raw_data["sig_shape"])
        if "sync_offset" in raw_data:
            data["sync_offset"] = raw_data["sync_offset"]
        return data


# stolen from hyperspy
def get_header_dtype_list(endianess='<'):
    end = endianess
    dtype_list = \
        [
            ('ID', (bytes, 6)),
            ('MAGIC', end + 'u2'),
            ('Data_offset_1', end + 'u4'),      # Offset VBF
            ('Data_offset_2', end + 'u4'),      # Offset DPs
            ('UNKNOWN1', end + 'u4'),           # Flags for ASTAR software?
            ('DP_SZ', end + 'u2'),              # Pixel dim DPs
            ('DP_rotation', end + 'u2'),        # [degrees ( * 100 ?)]
            ('NX', end + 'u2'),                 # Scan dim 1
            ('NY', end + 'u2'),                 # Scan dim 2
            ('Scan_rotation', end + 'u2'),      # [100 * degrees]
            ('SX', end + 'f8'),                 # Pixel size [nm]
            ('SY', end + 'f8'),                 # Pixel size [nm]
            ('Beam_energy', end + 'u4'),        # [V]
            ('SDP', end + 'u2'),                # Pixel size [100 * ppcm]
            ('Camera_length', end + 'u4'),      # [10 * mm]
            ('Acquisition_time', end + 'f8'),   # [Serial date]
        ] + [
            ('Centering_N%d' % i, 'f8') for i in range(8)
        ] + [
            ('Distortion_N%02d' % i, 'f8') for i in range(14)
        ]

    return dtype_list


class BloFileSet(FileSet):
    pass


class BloDataSet(DataSet):
    # FIXME include sample file for doctest, see Issue #86
    """
    Read Nanomegas .blo files

    Examples
    --------

    >>> ds = ctx.load("blo", path="/path/to/file.blo")  # doctest: +SKIP

    Parameters
    ----------
    path: str
        Path to the file

    endianess: str
        either '<' or '>' for little or big endian

    nav_shape: tuple of int, optional
        A n-tuple that specifies the size of the navigation region ((y, x), but
        can also be of length 1 for example for a line scan, or length 3 for
        a data cube, for example)

    sig_shape: tuple of int, optional
        Signal/detector size (height, width)

    sync_offset: int, optional
        If positive, number of frames to skip from start
        If negative, number of blank frames to insert at start
    """
    def __init__(self, path, tileshape=None, endianess='<', nav_shape=None,
                 sig_shape=None, sync_offset=0, io_backend=None):
        super().__init__(io_backend=io_backend)
        # handle backwards-compatability:
        if tileshape is not None:
            warnings.warn(
                "tileshape argument is ignored and will be removed after 0.6.0",
                FutureWarning
            )
        self._path = path
        self._header = None
        self._endianess = endianess
        self._shape = None
        self._filesize = None
        self._nav_shape = tuple(nav_shape) if nav_shape else nav_shape
        self._sig_shape = tuple(sig_shape) if sig_shape else sig_shape
        self._sync_offset = sync_offset

    def initialize(self, executor):
        self._header = h = executor.run_function(self._read_header)
        NY = int(h['NY'][0])
        NX = int(h['NX'][0])
        DP_SZ = int(h['DP_SZ'][0])
        self._image_count = NY * NX
        if self._nav_shape is None:
            self._nav_shape = (NY, NX)
        if self._sig_shape is None:
            self._sig_shape = (DP_SZ, DP_SZ)
        elif int(prod(self._sig_shape)) != (DP_SZ * DP_SZ):
            raise DataSetException(
                "sig_shape must be of size: %s" % (DP_SZ * DP_SZ)
            )
        self._nav_shape_product = int(prod(self._nav_shape))
        self._sync_offset_info = self.get_sync_offset_info()
        self._shape = Shape(self._nav_shape + self._sig_shape, sig_dims=len(self._sig_shape))
        self._meta = DataSetMeta(
            shape=self._shape,
            raw_dtype=np.dtype("u1"),
            sync_offset=self._sync_offset,
            image_count=self._image_count,
        )
        self._filesize = executor.run_function(self._get_filesize)
        return self

    @classmethod
    def detect_params(cls, path, executor):
        try:
            ds = cls(path, endianess='<')
            ds = ds.initialize(executor)
            if not executor.run_function(ds.check_valid):
                return False
            return {
                "parameters": {
                    "path": path,
                    "nav_shape": tuple(ds.shape.nav),
                    "sig_shape": tuple(ds.shape.sig),
                    "tileshape": (1, 8) + tuple(ds.shape.sig),
                    "endianess": "<",
                },
                "info": {
                    "image_count": int(prod(ds.shape.nav)),
                    "native_sig_shape": tuple(ds.shape.sig),
                }
            }
        except Exception:
            return False

    @classmethod
    def get_msg_converter(cls):
        return BLODatasetParams

    @classmethod
    def get_supported_extensions(cls):
        return {"blo"}

    @property
    def dtype(self):
        return self._meta.raw_dtype

    @property
    def shape(self):
        if self._shape is None:
            raise RuntimeError("please call initialize() before using the dataset")
        return self._shape

    def _read_header(self):
        # FIXME: do this read via the IO backend!
        with open(self._path, 'rb') as f:
            return np.fromfile(f, dtype=get_header_dtype_list(self._endianess), count=1)

    def _get_filesize(self):
        return os.stat(self._path).st_size

    @property
    def header(self):
        if self._header is None:
            raise RuntimeError("please call initialize() before using the dataset")
        return self._header

    def check_valid(self):
        try:
            header = self._read_header()
            magic = header['MAGIC'][0]
            if magic != MAGIC_EXPECT:
                raise DataSetException(f"invalid magic number: {magic:x} != {MAGIC_EXPECT:x}")
            return True
        except OSError as e:
            raise DataSetException("invalid dataset: %s" % e) from e

    def get_cache_key(self):
        return {
            "path": self._path,
            "endianess": self._endianess,
            "shape": tuple(self.shape),
            "sync_offset": self._sync_offset,
        }

    def _get_fileset(self):
        return BloFileSet([
            File(
                path=self._path,
                start_idx=0,
                end_idx=self._image_count,
                native_dtype=self._endianess + "u1",
                sig_shape=self.shape.sig,
                frame_header=6,
                file_header=int(self.header['Data_offset_2'][0]),
            )
        ], frame_header_bytes=6)

    def get_partitions(self):
        fileset = self._get_fileset()
        for part_slice, start, stop in self.get_slices():
            yield BasePartition(
                meta=self._meta,
                fileset=fileset,
                partition_slice=part_slice,
                start_frame=start,
                num_frames=stop - start,
                io_backend=self.get_io_backend(),
            )

    def __repr__(self):
        return f"<BloDataSet of {self.dtype} shape={self.shape}>"



import numpy as np

from libertem.analysis.sumfft import SumfftAnalysis
from libertem.io.dataset.memory import MemoryDataSet


def test_sum_fft_analysis_defaults(lt_ctx):
    data = np.zeros([3*3, 8, 8]).astype(np.float32)
    dataset = MemoryDataSet(data=data, tileshape=(1, 8, 8),
                            num_partitions=2, sig_dims=2)
    analysis = SumfftAnalysis(dataset=dataset, parameters={})
    lt_ctx.run(analysis)


def test_sum_fft_analysis_masked(lt_ctx):
    data = np.zeros([3*3, 8, 8]).astype(np.float32)
    dataset = MemoryDataSet(data=data, tileshape=(1, 8, 8),
                            num_partitions=2, sig_dims=2)
    analysis = SumfftAnalysis(dataset=dataset, parameters={
        'real_rad': 1,
        'real_centerx': 1,
        'real_centery': 1,
    })
    lt_ctx.run(analysis)



import io
import nbformat as nbf
from .code_template import CodeTemplate


class Notebook:
    '''
    Interface for creating jupyter notebook.
    '''
    def __init__(self):
        self.nb = nbf.v4.new_notebook()

    def add_code(self, code_):
        new_cell = nbf.v4.new_code_cell(code_)
        self.nb['cells'].append(new_cell)

    def add_doc(self, doc_):
        new_cell = nbf.v4.new_markdown_cell(doc_)
        self.nb['cells'].append(new_cell)

    def generate(self):
        f = io.StringIO()
        nbf.write(self.nb, f)
        return f


def notebook_generator(conn, dataset, comp, save=False):
    '''
    Generates a jupyter notebook file using the code
    generated by :class:`CodeTemplate`.

    For creating additional code/doc cell, generate
    code in :class:`CodeTemplate` and add cell here.
    '''
    # initialization
    nb = Notebook()
    instance = CodeTemplate(conn, dataset, comp)

    nb.add_code(instance.dependency())
    nb.add_code(instance.initial_setup())

    ctx, conn_docs = instance.connection()
    nb.add_doc(conn_docs)
    nb.add_code(ctx)
    nb.add_code(instance.dataset())

    for docs, analysis, plots, store in instance.analysis():
        nb.add_doc(docs)
        nb.add_code(analysis)
        for plot in plots:
            nb.add_code(plot)
        if save:
            nb.add_code(store)

    return nb.generate()



from collections import namedtuple
import itertools
from typing import Union
from numbers import Number

import dask
import numpy as np
import numpy.typing as nt


fake_np_flags = namedtuple('Flags', ['c_contiguous'])


class DaskInplaceWrapper:
    def __init__(self, dask_array):
        """
        Wraps a dask array providing methods such that array
        slicing can be done inplace even when slicing into a slice
        of the array. Works for simple slicing cases, i.e. slicing
        with slice(), integers, : and ...

        The first slice is set using DaskInplaceWrapper.set_slice(slice)
        while secondary slices are done using normal array[slice] syntax

        In the current usage, this wrapper is created and configured
        from DaskBufferWrapper._get_slice
        """
        assert isinstance(dask_array, dask.array.Array)
        self._array = dask_array
        self._slice = None

    @property
    def flags(self):
        return fake_np_flags(c_contiguous=True)

    @property
    def data(self):
        return self._array

    @property
    def dtype(self):
        return self.data.dtype

    @property
    def shape(self):
        return self.data.shape

    @property
    def size(self):
        return self.data.size

    @property
    def ndim(self):
        return self.data.ndim

    def check_valid_slices(self, slices):
        if not isinstance(slices, tuple):
            slices = (slices,)
        assert all(isinstance(s, (slice, int)) or s in (Ellipsis, None) for s in slices)
        assert len(slices) <= self.ndim

    def set_slice(self, slices):
        self.check_valid_slices(slices)
        self._slice = slices

    def clear_slice(self):
        self._slice = None

    def unwrap_sliced(self):
        """
        Unwrap the array as a sliced view, if self._slice is set
        """
        if self._slice is None:
            return self._array
        else:
            return self._array[self._slice]

    def __getitem__(self, key):
        if self._slice is None:
            return self._array[key]
        else:
            combined_slice = combine_slices_multid(self._slice, key, self._array.shape)
            return self._array[combined_slice]

    def __setitem__(self, key, value: Union[nt.ArrayLike, Number]):
        if not np.isscalar(value) and value.size == 1:
            # Avoids deprecated Numpy behaviour of implicitly
            # extracting a scalar from an array during assignment
            value = value.item()
        try:
            if self._slice is None:
                self._array[key] = value
            else:
                combined_slice = combine_slices_multid(self._slice, key, self._array.shape)
                self._array[combined_slice] = value
        except NotImplementedError:
            raise NotImplementedError(
                "Assignment into Dask array failed. This feature requires "
                f"Dask version >= 2021.4.1. Installed is {dask.__version__}."
            )


def combine_slices_multid(slices1, slices2, shape):
    null_slice = slice(None, None, None)

    if not isinstance(slices1, tuple):
        slices1 = (slices1,)
    if not isinstance(slices2, tuple):
        slices2 = (slices2,)

    if all([s == null_slice for s in slices2]):
        return slices1

    if all([s == null_slice for s in slices1]):
        return slices2

    # Handle the Ellipsis case for slice1
    # can handle Ellipsis leading, central and training
    # in slices 1 because it is the primary slice
    # This will pad slice1 to be the same size as shape
    if Ellipsis in slices1:
        assert sum(e is Ellipsis for e in slices1) == 1
        el_idx = slices1.index(Ellipsis)
        before = slices1[:el_idx]
        after = slices1[el_idx + 1:]
        to_add = len(shape) - (len(before) + (len(after)))
        _null = (null_slice,) * to_add
        slices1 = before + _null + after

    # Handle the Ellipsis case for slice2
    # need to work on this to handle leading or central Ellipsis
    # trailing ellipsis is supported as the padding is implemented after
    if Ellipsis in slices2:
        if Ellipsis != slices2[-1] or sum(e is Ellipsis for e in slices1) > 1:
            raise IndexError("No support for Ellipsis in subslice except at end")
        slices2 = slices2[:-1]

    # Must pad slices1 to length of array itself so that
    # we can use it as a reference for padding slices2
    while len(slices1) < len(shape):
        slices1 = slices1 + (null_slice,)

    # If slice1 contains integers we must pad slices2 for combining
    # This section of code is not complete/robust, there are cases
    # where slices1 is all non-integer but slices2 is not of the same length!
    # e.g. slices1 = (slice(0, 64), slice(0, 64)) and slices2 = (slice(0, 32),)
    # need to have logic for slicing only the first dimension here
    s2_gen = iter(slices2)
    _slices2 = []
    for s1 in slices1:
        if isinstance(s1, int):
            _slices2.append(null_slice)
        else:
            _slices2.append(next(s2_gen, null_slice))
    next_el = next(s2_gen, None)
    assert next_el is None
    slices2 = tuple(_slices2)

    combined_slices = []
    for _slice1, _slice2, _dimension in itertools.zip_longest(slices1, slices2, shape,
                                                              fillvalue=null_slice):
        combined_slice = combine_slices(_slice1, _slice2, _dimension)
        combined_slices.append(combined_slice)
    return tuple(combined_slices)


def combine_slices(slice1, slice2, length):
    """
    https://stackoverflow.com/a/26783035
    Modified to support slices which are just element selection
    i.e. an integer for slice1 or slice2


    returns a slice that is a combination of the two slices.
    As in
      x[slice1][slice2]
    becomes
      combined_slice = slice_combine(slice1, slice2, len(x))
      x[combined_slice]

    :param slice1: The first slice
    :param slice2: The second slice
    :param length: The length of the first dimension of data being sliced. (eg len(x))
    """
    if isinstance(slice1, int):
        # by definition we can't slice further on this dimension
        assert slice2 is None or slice2 == slice(None, None, None)
        return slice1

    # First get the step sizes of the two slices.
    slice1_step = (slice1.step if slice1.step is not None else 1)

    # Use slice1.indices to get the actual indices returned from slicing with slice1
    slice1_indices = slice1.indices(length)

    if isinstance(slice2, int):
        # Represents a single element from array[slice1]
        return range(*slice1_indices)[slice2]

    # From this point we are combining two non-integer slices
    slice2_step = (slice2.step if slice2.step is not None else 1)

    # The final step size
    step = slice1_step * slice2_step

    # We calculate the length of the first slice
    slice1_length = (abs(slice1_indices[1] - slice1_indices[0]) - 1) // abs(slice1_indices[2])

    # If we step in the same direction as the start,stop, we get at least one datapoint
    if (slice1_indices[1] - slice1_indices[0]) * slice1_step > 0:
        slice1_length += 1
    else:
        # Otherwise, The slice is zero length.
        return slice(0, 0, step)

    # Use the length after the first slice to get the indices returned from a
    # second slice starting at 0.
    slice2_indices = slice2.indices(slice1_length)

    # if the final range length = 0, return
    if not (slice2_indices[1] - slice2_indices[0]) * slice2_step > 0:
        return slice(0, 0, step)

    # We shift slice2_indices by the starting index in slice1 and the
    # step size of slice1
    start = slice1_indices[0] + slice2_indices[0] * slice1_step
    stop = slice1_indices[0] + slice2_indices[1] * slice1_step

    # slice.indices will return -1 as the stop index when slice.stop should be set to None.
    if start > stop:
        if stop < 0:
            stop = None

    return slice(start, stop, step)



import sparse
import pytest
import numpy as np

from libertem.common.exceptions import UDFException
from libertem.io.corrections import CorrectionSet
from libertem.io.dataset.base import Negotiator
from libertem.io.dataset.memory import MemoryDataSet
from libertem.udf.base import UDF, NoOpUDF

from utils import _mk_random


def test_scale_factors():
    neg = Negotiator()
    # scenario: k2is.
    assert neg._get_scale_factors(
        shape=(930, 16),
        containing_shape=(1860, 2048),
        size=1024,
    ) == [1, 1]

    assert neg._get_scale_factors(
        shape=(930, 16),
        containing_shape=(1860, 2048),
        size=930 * 16 * 16
    ) == [2, 8]

    assert neg._get_scale_factors(
        shape=(930, 16),
        containing_shape=(1860, 2048),
        size=930 * 16 * 128
    ) == [2, 64]

    # slightly above, but not enough to fit another block: we err on the small side
    assert neg._get_scale_factors(
        shape=(930, 16),
        containing_shape=(1860, 2048),
        size=930 * 16 * 129
    ) == [2, 64]

    # larger size than we can accomodate with our containing shape:
    assert neg._get_scale_factors(
        shape=(930, 16),
        containing_shape=(1860, 2048),
        size=1860 * 2048 * 2
    ) == [2, 128]


class TilingUDF(UDF):
    def process_tile(self, tile):
        pass

    def get_tiling_preferences(self):
        return {
            "depth": UDF.TILE_DEPTH_DEFAULT,
            "total_size": UDF.TILE_SIZE_MAX,
        }


class TilingUDFBestFit(UDF):
    def process_tile(self, tile):
        pass

    def get_tiling_preferences(self):
        return {
            "depth": UDF.TILE_DEPTH_DEFAULT,
            "total_size": UDF.TILE_SIZE_BEST_FIT,
        }


class TilingUDFFrame(UDF):
    def process_frame(self, frame):
        pass


class TilingUDFPartition(UDF):
    def process_partition(self, partition):
        pass


def test_get_scheme_tile(default_raw):
    neg = Negotiator()
    p = next(default_raw.get_partitions())
    udf = TilingUDFBestFit()
    scheme = neg.get_scheme(
        udfs=[udf],
        dataset=default_raw,
        approx_partition_shape=p.shape,
        read_dtype=np.float32,
        roi=None
    )
    assert scheme.shape.sig.dims == 2
    print(neg._get_udf_size_pref(udf))
    print(scheme._debug)
    print(p.shape)
    assert tuple(scheme.shape) == (64, 32, 128)


def test_get_scheme_frame(default_raw):
    neg = Negotiator()
    p = next(default_raw.get_partitions())
    udf = TilingUDFFrame()
    scheme = neg.get_scheme(
        udfs=[udf],
        dataset=default_raw,
        approx_partition_shape=p.shape,
        read_dtype=np.float32,
        roi=None
    )
    assert scheme.shape.sig.dims == 2
    assert tuple(scheme.shape) == (16, 128, 128)


def test_get_scheme_partition(default_raw):
    neg = Negotiator()
    p = next(default_raw.get_partitions())
    udf = TilingUDFPartition()
    scheme = neg.get_scheme(
        udfs=[udf],
        dataset=default_raw,
        approx_partition_shape=p.shape,
        read_dtype=np.float32,
        roi=None
    )
    assert scheme.shape.sig.dims == 2
    assert tuple(scheme.shape) == (128, 128, 128)


def test_get_scheme_upper_size_1():
    """
    Test that will hit the 2**20 default size
    """
    data = _mk_random(size=(1024, 144, 144))
    dataset = MemoryDataSet(
        data=data,
        num_partitions=1,
        sig_dims=2
    )

    neg = Negotiator()
    p = next(dataset.get_partitions())
    udf = TilingUDFBestFit()
    scheme = neg.get_scheme(
        udfs=[udf],
        dataset=dataset,
        approx_partition_shape=p.shape,
        read_dtype=np.float32,
        roi=None
    )
    assert scheme.shape.sig.dims == 2
    assert tuple(scheme.shape) == (65, 28, 144)


@pytest.mark.xfail(
    reason="With global TilingScheme we can't handle this as before,"
           " will be fixed with #382",
    raises=AssertionError,
)
def test_get_scheme_upper_size_roi():
    """
    Confirm that a small ROI will not be split
    up unnecessarily.
    """
    data = _mk_random(size=(1024, 144, 144))
    dataset = MemoryDataSet(
        data=data,
        num_partitions=1,
        sig_dims=2
    )

    roi = np.zeros(dataset.shape.nav, dtype=bool)
    # All in a single partition here
    roi[0] = True
    roi[512] = True
    roi[-1] = True

    neg = Negotiator()
    p = next(dataset.get_partitions())
    udf = TilingUDFBestFit()
    scheme = neg.get_scheme(
        udfs=[udf],
        approx_partition_shape=p.shape,
        dataset=dataset,
        read_dtype=np.float32,
        roi=roi
    )
    assert scheme.shape.sig.dims == 2
    assert tuple(scheme.shape) == (3, 144, 144)


def test_get_scheme_upper_size_2():
    """
    Test that will hit the 2**20 default size
    """
    data = _mk_random(size=(2048, 264, 264))
    dataset = MemoryDataSet(
        data=data,
        num_partitions=1,
        sig_dims=2,
        base_shape=(1, 8, 264),
    )

    neg = Negotiator()
    p = next(dataset.get_partitions())
    udf = TilingUDFBestFit()
    scheme = neg.get_scheme(
        udfs=[udf],
        approx_partition_shape=p.shape,
        dataset=dataset,
        read_dtype=np.float32,
        roi=None
    )
    assert scheme.shape.sig.dims == 2
    assert tuple(scheme.shape) == (124, 8, 264)


class UDFWithLargeDepth(UDF):
    def process_tile(self, tile):
        pass

    def get_tiling_preferences(self):
        return {
            "depth": 128,
            "total_size": UDF.TILE_SIZE_BEST_FIT,
        }


class UDFUnlimitedDepth(UDF):
    def process_tile(self, tile):
        pass

    def get_tiling_preferences(self):
        return {
            "depth": UDF.TILE_DEPTH_MAX,
            "total_size": UDF.TILE_SIZE_MAX,
        }


def test_limited_depth():
    data = _mk_random(size=(32, 1860, 2048))
    dataset = MemoryDataSet(
        data=data,
        num_partitions=1,
        sig_dims=2,
        base_shape=(1, 930, 16),
        force_need_decode=True,
    )

    neg = Negotiator()
    p = next(dataset.get_partitions())
    udf = UDFWithLargeDepth()
    scheme = neg.get_scheme(
        udfs=[udf],
        approx_partition_shape=p.shape,
        dataset=dataset,
        read_dtype=np.float32,
        roi=None
    )
    print(scheme._debug)
    assert scheme._debug["need_decode"]
    assert scheme.shape.sig.dims == 2
    assert tuple(scheme.shape) == (17, 930, 16)


def test_correction_size_overflow():
    data = _mk_random(size=(32, 1860, 2048))
    dataset = MemoryDataSet(
        data=data,
        num_partitions=1,
        sig_dims=2,
        base_shape=(1, 930, 16),
        force_need_decode=True,
    )

    neg = Negotiator()
    p = next(dataset.get_partitions())
    udf = UDFWithLargeDepth()

    excluded_coords = np.array([
        (930, ),
        (16, )
    ])
    excluded_pixels = sparse.COO(coords=excluded_coords, shape=dataset.shape.sig, data=True)
    corr = CorrectionSet(excluded_pixels=excluded_pixels)

    scheme = neg.get_scheme(
        udfs=[udf], approx_partition_shape=p.shape,
        dataset=dataset,
        read_dtype=np.float32, roi=None,
        corrections=corr,
    )
    print(scheme._debug)
    assert scheme._debug["need_decode"]
    assert scheme.shape.sig.dims == 2
    assert tuple(scheme.shape) == (4, 1860, 32)


def test_depth_max_size_max():
    data = _mk_random(size=(32, 1860, 2048))
    dataset = MemoryDataSet(
        data=data,
        num_partitions=1,
        sig_dims=2,
        base_shape=(1, 930, 16),
        force_need_decode=False,
    )

    neg = Negotiator()
    p = next(dataset.get_partitions())
    udf = UDFUnlimitedDepth()
    scheme = neg.get_scheme(
        udfs=[udf],
        approx_partition_shape=p.shape,
        dataset=dataset,
        read_dtype=np.float32,
        roi=None
    )
    print(scheme._debug)
    assert not scheme._debug["need_decode"]
    assert scheme.shape.sig.dims == 2
    assert tuple(scheme.shape) == (32, 1860, 2048)


def test_multi_by_frame_wins():
    by_frame = TilingUDFFrame()
    other_unlimited = UDFUnlimitedDepth()
    other_best_fit = TilingUDFBestFit()
    other_deep = UDFWithLargeDepth()

    udfs = [
        by_frame,
        other_unlimited,
        other_best_fit,
        other_deep,
    ]

    data = _mk_random(size=(32, 1860, 2048))
    dataset = MemoryDataSet(
        data=data,
        num_partitions=1,
        sig_dims=2,
        base_shape=(1, 930, 16),
        force_need_decode=False,
    )

    neg = Negotiator()
    p = next(dataset.get_partitions())
    scheme = neg.get_scheme(
        udfs=udfs,
        dataset=dataset,
        approx_partition_shape=p.shape,
        read_dtype=np.float32,
        roi=None
    )
    print(scheme._debug)
    assert scheme.shape.sig.dims == 2
    assert tuple(scheme.shape) == (1, 1860, 2048)


def test_multi_no_by_frame_small_size_wins():
    other_unlimited = UDFUnlimitedDepth()
    other_best_fit = TilingUDFBestFit()
    other_deep = UDFWithLargeDepth()

    udfs = [
        other_unlimited,
        other_best_fit,
        other_deep,
    ]

    data = _mk_random(size=(32, 1860, 2048))
    dataset = MemoryDataSet(
        data=data,
        num_partitions=1,
        sig_dims=2,
        base_shape=(1, 930, 16),
        force_need_decode=False,
    )

    neg = Negotiator()
    p = next(dataset.get_partitions())
    scheme = neg.get_scheme(
        udfs=udfs,
        dataset=dataset,
        approx_partition_shape=p.shape,
        read_dtype=np.float32,
        roi=None
    )
    print(scheme._debug)
    assert scheme.shape.sig.dims == 2
    assert tuple(scheme.shape) == (17, 930, 16)


def test_multi_partition_wins():
    # FIXME: the constellatin partition+tile or partition+frame or similar
    # can be optimized by using a fitting tiling scheme for tile/frame processing
    # and accumulating the whole partition into a buffer, then running process_partition
    # after the tile loop.
    other_unlimited = UDFUnlimitedDepth()
    other_best_fit = TilingUDFBestFit()
    other_deep = UDFWithLargeDepth()
    udf_partition = TilingUDFPartition()

    udfs = [
        udf_partition,
        other_unlimited,
        other_best_fit,
        other_deep,
    ]

    data = _mk_random(size=(32, 1860, 2048))
    dataset = MemoryDataSet(
        data=data,
        num_partitions=1,
        sig_dims=2,
        base_shape=(1, 930, 16),
        force_need_decode=False,
    )

    neg = Negotiator()
    p = next(dataset.get_partitions())
    scheme = neg.get_scheme(
        udfs=udfs,
        approx_partition_shape=p.shape,
        dataset=dataset,
        read_dtype=np.float32,
        roi=None
    )
    print(scheme._debug)
    assert scheme.shape.sig.dims == 2
    assert tuple(scheme.shape) == (32, 1860, 2048)


def test_base_shape_adjustment_asis(lt_ctx_fast):
    '''
    Confirm that dataset can just accept tile shape as-is
    '''
    class MockDectrisMemoryDataSet(MemoryDataSet):
        def get_max_io_size(self):
            return 12*512*512*8

        def adjust_tileshape(self, tileshape, roi):
            return tileshape

    ds = MockDectrisMemoryDataSet(
        datashape=(3, 3, 512, 512),
    )

    bad_y = (168, 291, 326, 301, 343, 292,   0,   0,   0,   0,   0, 511)
    bad_x = (496, 458, 250, 162, 426, 458, 393, 396, 413, 414, 342, 491)

    corr = CorrectionSet(
        excluded_pixels=sparse.COO(
            coords=(bad_y, bad_x),
            data=1,
            shape=ds.shape.sig
        )
    )
    lt_ctx_fast.run_udf(dataset=ds, udf=NoOpUDF(), corrections=corr)


def test_base_shape_adjustment_valid(lt_ctx_fast):
    '''
    This emulates an issue with the DECTRIS acquisition of LiberTEM Live
    that overrides the tile shape to full frames.
    Bad pixel patching leads to a base shape that is not evenly divinding the frame shape,
    which then led to a conflict with the overridden shape.

    Solution was to always accept full frames.
    '''
    class MockDectrisMemoryDataSet(MemoryDataSet):
        def get_max_io_size(self):
            return 12*512*512*8

        def adjust_tileshape(self, tileshape, roi):
            depth = 12
            assert tileshape[0] != depth
            return (depth, *self.meta.shape.sig)

    ds = MockDectrisMemoryDataSet(
        datashape=(3, 3, 512, 512),
    )

    bad_y = (168, 291, 326, 301, 343, 292,   0,   0,   0,   0,   0, 511)
    bad_x = (496, 458, 250, 162, 426, 458, 393, 396, 413, 414, 342, 491)

    corr = CorrectionSet(
        excluded_pixels=sparse.COO(
            coords=(bad_y, bad_x),
            data=1,
            shape=ds.shape.sig
        )
    )
    lt_ctx_fast.run_udf(dataset=ds, udf=NoOpUDF(), corrections=corr)


def test_base_shape_adjustment_invalid(lt_ctx_fast):
    '''
    Forbid arbitrary tile shape adjustments by dataset if pixel patching is active.

    Only original tile sig shape and full frames are allowed.
    '''
    class MockDectrisMemoryDataSet(MemoryDataSet):
        def get_max_io_size(self):
            return 12*512*512*8

        def adjust_tileshape(self, tileshape, roi):
            return (tileshape[0], 234, 345)

    ds = MockDectrisMemoryDataSet(
        datashape=(3, 3, 512, 512),
    )

    bad_y = (168, 291, 326, 301, 343, 292,   0,   0,   0,   0,   0, 511)
    bad_x = (496, 458, 250, 162, 426, 458, 393, 396, 413, 414, 342, 491)

    corr = CorrectionSet(
        excluded_pixels=sparse.COO(
            coords=(bad_y, bad_x),
            data=1,
            shape=ds.shape.sig
        )
    )
    with pytest.raises(ValueError):
        lt_ctx_fast.run_udf(dataset=ds, udf=NoOpUDF(), corrections=corr)


class BadGetMethodUDF(UDF):
    def get_method(self):
        return 42


def test_unrecognized_udf_method(default_raw):
    p = next(default_raw.get_partitions())
    with pytest.raises(UDFException):
        Negotiator().get_scheme(
            udfs=[BadGetMethodUDF()],
            dataset=default_raw,
            approx_partition_shape=p.shape,
            read_dtype=np.float32,
            roi=None
        )


def test_missing_udfs_get_intent():
    with pytest.raises(ValueError):
        Negotiator()._get_intent(
            udfs=[],
        )



import json
import asyncio

import pytest
import websockets

from utils import assert_msg
from aio_utils import (
    create_connection, create_analysis,
    create_update_compound_analysis, create_job_for_analysis,
)


pytestmark = [pytest.mark.web_api]


def _get_ds_params():
    return {
        "dataset": {
            "params": {
                "type": "memory",
                "tileshape": [7, 32, 32],
                "datashape": [256, 32, 32],
                "tiledelay": 0.1,
                "num_partitions": 16,
                "sync_offset": 0,
            }
        }
    }


@pytest.mark.asyncio
async def test_cancel_udf_job(
    base_url, default_raw, http_client, server_port, shared_state, local_cluster_url, default_token,
):
    await create_connection(base_url, http_client, local_cluster_url, default_token)

    print("checkpoint 1")

    # connect to ws endpoint:
    ws_url = f"ws://127.0.0.1:{server_port}/api/events/?token={default_token}"
    async with websockets.connect(ws_url) as ws:
        print("checkpoint 2")
        initial_msg = json.loads(await ws.recv())
        assert_msg(initial_msg, 'INITIAL_STATE')

        ds_uuid = "ae5d23bd-1f2a-4c57-bab2-dfc59a1219f3"
        ds_url = "{}/api/datasets/{}/?token={}".format(
            base_url, ds_uuid, default_token,
        )

        ds_data = _get_ds_params()
        async with http_client.put(ds_url, json=ds_data) as resp:
            assert resp.status == 200
            resp_json = await resp.json()
            assert_msg(resp_json, 'CREATE_DATASET')

        # same msg via ws:
        msg = json.loads(await ws.recv())
        assert_msg(msg, 'CREATE_DATASET')

        ca_uuid, ca_url = await create_update_compound_analysis(
            ws, http_client, base_url, ds_uuid, token=default_token,
        )

        analysis_uuid, analysis_url = await create_analysis(
            ws, http_client, base_url, ds_uuid, ca_uuid, token=default_token,
        )

        job_uuid, job_url = await create_job_for_analysis(
            ws, http_client, base_url, analysis_uuid, token=default_token,
        )

        await asyncio.sleep(0)  # for debugging, set to >0

        async with http_client.delete(job_url) as resp:
            assert resp.status == 200
            assert_msg(await resp.json(), 'CANCEL_JOB')

        # wait for CANCEL_JOB message:
        done = False
        num_seen = 0
        while not done:
            msg = json.loads(await ws.recv())
            if msg['messageType'] != "JOB_PROGRESS":
                num_seen += 1
            if msg['messageType'] == 'TASK_RESULT':
                assert_msg(msg, 'TASK_RESULT')
                assert msg['job'] == job_uuid
            elif msg['messageType'] == 'JOB_PROGRESS':
                assert msg['job'] == job_uuid
            elif msg['messageType'] == 'CANCEL_JOB':
                # this is the confirmation sent from the DELETE method handler:
                assert_msg(msg, 'CANCEL_JOB')
                assert msg['job'] == job_uuid
                done = True
            else:
                raise Exception("invalid message type: {}".format(msg['messageType']))
            if 'followup' in msg:
                for i in range(msg['followup']['numMessages']):
                    # drain binary messages:
                    msg = await ws.recv()

        assert num_seen < 4
        assert job_uuid not in shared_state.job_state.jobs

        # now we drain messages from websocket and look for CANCEL_JOB_DONE msg:
        done = False
        num_seen = 0
        types_seen = []
        msgs = []
        while not done:
            msg = json.loads(await ws.recv())
            print(msg)
            num_seen += 1
            types_seen.append(msg['messageType'])
            msgs.append(msg)
            if msg['messageType'] == 'TASK_RESULT':
                assert_msg(msg, 'TASK_RESULT')
                assert msg['job'] == job_uuid
            elif msg['messageType'] == 'JOB_PROGRESS':
                assert msg['job'] == job_uuid
            elif msg['messageType'] == 'CANCEL_JOB_DONE':
                assert msg['job'] == job_uuid
                done = True
            elif msg['messageType'] == 'FINISH_JOB':
                done = True  # we should not get this one...
            elif msg['messageType'] == 'JOB_ERROR':
                raise Exception('JOB_ERROR: {}'.format(msg['msg']))
            else:
                raise Exception("invalid message type: {}".format(msg['messageType']))
            if 'followup' in msg:
                for i in range(msg['followup']['numMessages']):
                    # drain binary messages:
                    msg = await ws.recv()
        for mtype in types_seen:
            assert mtype in {"CANCEL_JOB_DONE", "JOB_PROGRESS"}
        msgs_without_progress = [
            msg
            for msg in msgs
            if msg['messageType'] != "JOB_PROGRESS"
        ]
        assert len(msgs_without_progress) < 4

        # get rid of the dataset:
        async with http_client.delete(ds_url) as resp:
            assert resp.status == 200
            resp_json = await resp.json()
            assert_msg(resp_json, 'DELETE_DATASET')



import io
import os

import pytest
import numpy as np
import nbformat
from temp_utils import _get_hdf5_params
from libertem.web.notebook_generator.notebook_generator import notebook_generator
from nbconvert.preprocessors import ExecutePreprocessor


@pytest.mark.slow
def test_radial_fourier_default(hdf5_ds_2, tmpdir_factory, lt_ctx, local_cluster_url):
    datadir = tmpdir_factory.mktemp('template_tests')

    conn = {'connection': {'type': 'tcp', 'address': local_cluster_url}}
    path = hdf5_ds_2.path
    dataset = _get_hdf5_params(path)

    analysis = [{
            "analysisType": "RADIAL_FOURIER",
            "parameters": {
                    'shape': 'radial_fourier',
                    'cx': 0,
                    'cy': 0,
                    'ri': 0,
                    'ro': 2,
                    'n_bins': 2,
                    'max_order': 7,
                    }
    }]

    notebook = notebook_generator(conn, dataset, analysis, save=True)
    notebook = io.StringIO(notebook.getvalue())
    nb = nbformat.read(notebook, as_version=4)
    ep = ExecutePreprocessor(timeout=600)
    ep.preprocess(nb, {"metadata": {"path": datadir}})
    channels = [
        "absolute_0_0",
        "absolute_0_1"
    ]
    results = {}
    for channel in channels:
        data_path = os.path.join(datadir, f"radial_result_{channel}.npy")
        results[channel] = np.load(data_path)

    analysis = lt_ctx.create_radial_fourier_analysis(
                                        dataset=hdf5_ds_2,
                                        cx=0,
                                        cy=0,
                                        ri=0,
                                        ro=2,
                                        n_bins=2,
                                        max_order=7
                                )
    expected = lt_ctx.run(analysis)

    assert np.allclose(results["absolute_0_0"], expected["absolute_0_0"].raw_data)
    assert np.allclose(results["absolute_0_1"], expected["absolute_0_1"].raw_data)



import types
import warnings

import dask
import dask.delayed
import dask.distributed as dd

from .dask import DaskJobExecutor
from .concurrent import ConcurrentJobExecutor
from .inline import InlineJobExecutor


def get_dask_integration_executor():
    '''
    Query the current Dask scheduler and return a :class:`~libertem.common.executor.JobExecutor`
    that is compatible with it. See https://docs.dask.org/en/stable/scheduling.html
    for the meaning of the different scheduler types.

    This can be used to integrate LiberTEM in an existing Dask workflow. This
    may not achieve optimal LiberTEM performance and will usually not allow GPU
    processing with LiberTEM, but avoids potential compatibility issues from
    changing or duplicating the Dask scheduler in an existing workflow.

    .. versionadded:: 0.9.0

    If a :code:`dask.distributed.Client` is set as the scheduler, return a
    :class:`~libertem.executor.dask.DaskJobExecutor` using this :code:`Client`.

    If the Dask scheduler is :code:`'threads'`, return a
    :class:`~libertem.executor.concurrent.ConcurrentJobExecutor` backed
    by the same thread pool as used by Dask.

    If the Dask scheduler is :code:`'synchronous'`, return an
    :class:`~libertem.executor.inline.InlineJobExecutor`
    which mimics the single-process, single-thread behaviour of Dask.
    '''
    item = dask.delayed(1)
    dask_scheduler = dask.base.get_scheduler(collections=[item])
    # We first test for circumstances where we know how to return an adapted
    # JobExecutor instance.
    if isinstance(dask_scheduler, types.MethodType):
        if isinstance(dask_scheduler.__self__, dd.Client):
            # See https://github.com/dask/distributed/issues/6776
            if dask.config.get("distributed.worker.profile.enabled"):
                warnings.warn(
                    "Dask profiling seems to be enabled, which is known to cause issues with "
                    "the DM reader. "
                    "See https://github.com/dask/distributed/issues/6776"
                )
            return DaskJobExecutor(client=dask_scheduler.__self__)
    elif dask_scheduler is dask.threaded.get:
        if dask.threaded.default_pool:
            return ConcurrentJobExecutor(client=dask.threaded.default_pool)
    # ConcurrentJobExecutor is currently incompatible with ProcessPoolExecutor
    # since it can't pickle local functions.
    # Therefore, fall through to default case for now.
    # elif dask_scheduler is dask.multiprocessing.get:
    #     # FIXME more research needed if the pool is cached and if
    #     # one can get hold of it.
    #     return ConcurrentJobExecutor(
    #         client=concurrent.futures.ProcessPoolExecutor(),
    #         is_local=True
    #     )
    elif dask_scheduler is dask.local.get_sync:
        return InlineJobExecutor()

    # If we didn't return yet,
    # we fall through to the default case.
    return ConcurrentJobExecutor.make_local()



import logging
from contextlib import contextmanager
import os
import warnings
import threading

# NOTE: most imports are performed locally in the functions, to
# make sure these functions are usable as early as possible in the
# start-up of libertem (i.e. before numpy/numba/... are imported).
# In particular, that makes sure the environment variables can be set
# with set_num_threads_env() before the libraries are loaded.


log = logging.getLogger(__name__)
lock = threading.Lock()

__has_pyfftw = None
__has_pyfftw_threads = None
__has_pytorch = None

__pyfftw_counter = 0
__pyfftw_threads = None

__mitigation_counter = 0
__pyfftw_cached = None

__numba_counter = 0
__numba_threads = None

__pytorch_counter = 0
__pytorch_threads = None

# Make sure we don't load threadpoolctl when importing,
# just to be sure no C library gets loaded!
__threadpool_counter = 0
__threadpool_limiter = None
__threadpool_wrapper = None


def detect_pyfftw():
    global __has_pyfftw
    global __has_pyfftw_threads
    with lock:
        if __has_pyfftw is None:
            try:
                import pyfftw
                __has_pyfftw = True
                try:
                    pyfftw_threads = pyfftw.config.NUM_THREADS
                    pyfftw.config.NUM_THREADS = 2
                    data = pyfftw.empty_aligned((4, 4))
                    pyfftw.interfaces.numpy_fft.fft2(data)
                    __has_pyfftw_threads = True
                except ValueError:
                    __has_pyfftw_threads = False
                finally:
                    pyfftw.config.NUM_THREADS = pyfftw_threads
            except ImportError:
                __has_pyfftw = False


@contextmanager
def set_fftw_threads(n):
    global __has_pyfftw
    global __has_pyfftw_threads
    global __pyfftw_counter
    global __pyfftw_threads

    detect_pyfftw()

    try:
        with lock:
            if __has_pyfftw and __has_pyfftw_threads and __pyfftw_counter == 0:
                import pyfftw  # noqa:F811
                __pyfftw_threads = pyfftw.config.NUM_THREADS
                pyfftw.config.NUM_THREADS = n
            __pyfftw_counter += 1
        yield
    finally:
        with lock:
            __pyfftw_counter -= 1
            if (__has_pyfftw and __has_pyfftw_threads and __pyfftw_counter == 0
                    and __pyfftw_threads is not None):
                import pyfftw  # noqa:F811
                pyfftw.config.NUM_THREADS = __pyfftw_threads
                __pyfftw_threads = None


def detect_pytorch():
    global __has_pytorch

    with lock:
        if __has_pytorch is None:
            try:
                import torch  # noqa:F811,F401
                __has_pytorch = True
            except ImportError:
                __has_pytorch = False


@contextmanager
def set_torch_threads(n):
    global __has_pytorch
    global __pytorch_counter
    global __pytorch_threads

    detect_pytorch()

    try:
        with lock:
            if __has_pytorch and __pytorch_counter == 0:
                import torch  # noqa:F811
                __pytorch_threads = torch.get_num_threads()
                # See also https://pytorch.org/docs/stable/torch.html#parallelism
                # At the time of writing the difference between threads and interop threads
                # wasn't clear from the documentation. However, changing the
                # interop_threads on runtime
                # caused errors, so it is commented out here.
                # torch_interop_threads = torch.get_num_interop_threads()
                torch.set_num_threads(n)
                # torch.set_num_interop_threads(n)
            __pytorch_counter += 1
        yield
    finally:
        with lock:
            __pytorch_counter -= 1
            if __has_pytorch and __pytorch_counter == 0 and __pytorch_threads is not None:
                import torch
                torch.set_num_threads(__pytorch_threads)
                # torch.set_num_interop_threads(torch_interop_threads)
                __pytorch_threads = None


@contextmanager
def set_numba_threads(n):
    global __numba_counter
    global __numba_threads
    import numba
    from numba.core.config import NUMBA_NUM_THREADS

    try:
        with lock:
            if __numba_counter == 0:
                __numba_threads = numba.get_num_threads()
                if n > NUMBA_NUM_THREADS:
                    warnings.warn(
                        f"Attempting to set threads to {n}, which is larger than "
                        f"NUMBA_NUM_THREADS={NUMBA_NUM_THREADS}. "
                        f"Setting to allowed maximum NUMBA_NUM_THREADS instead."
                    )
                n = min(n, NUMBA_NUM_THREADS)
                numba.set_num_threads(n)
            __numba_counter += 1
        yield
    finally:
        with lock:
            __numba_counter -= 1
            if __numba_counter == 0 and __numba_threads is not None:
                numba.set_num_threads(__numba_threads)
                __numba_threads = None


class ThreadpoolWrapper:
    def __init__(self, repeats=2):
        import threadpoolctl
        self._info = threadpoolctl.threadpool_info()
        self._controller = threadpoolctl.ThreadpoolController()
        self._stable = 0
        self.repeats = repeats

    def check_update(self):
        import threadpoolctl
        new_info = threadpoolctl.threadpool_info()
        if new_info != self._info:
            self._stable = 0
            self._controller = threadpoolctl.ThreadpoolController()
            self._info = new_info
        else:
            self._stable += 1

    # Currently not used due to executor overhead
    # def reset(self):
    #     self ._stable = 0

    def __call__(self, limits):
        if self._stable < self.repeats:
            self.check_update()
        return self._controller.limit(limits=limits)


# Currently not used due to executor overhead
# def reset_module_cache():
#     if __threadpool_wrapper is not None:
#         __threadpool_wrapper.reset()


@contextmanager
def set_num_threads(n):
    # Make sure modules that use BLAS are loaded
    import scipy  # noqa: F401
    import numpy as np  # noqa: F401
    global __threadpool_wrapper
    global __threadpool_counter
    global __threadpool_limiter

    with lock:
        if __threadpool_wrapper is None:
            __threadpool_wrapper = ThreadpoolWrapper()
    # We use __threadpool_wrapper last so that it can cover
    # libraries that the other ones load
    with set_fftw_threads(n):
        with set_torch_threads(n):
            with set_numba_threads(n):
                try:
                    with lock:
                        if __threadpool_counter == 0:
                            # The limit is set upon calling ThreadpoolController.limit()
                            __threadpool_limiter = __threadpool_wrapper(n)
                        __threadpool_counter += 1
                    yield
                finally:
                    with lock:
                        __threadpool_counter -= 1
                        if __threadpool_counter == 0 and __threadpool_limiter is not None:
                            __threadpool_limiter.restore_original_limits()
                            __threadpool_limiter = None


@contextmanager
def set_num_threads_env(n=1, set_numba=None):
    """Set the maximum number of threads via environment variables.
    Currently sets variables for MKL, OMP, OPENBLAS and NUMBA.

    This is needed if you want to limit the number of threads that
    are created at startup of those libraries, because the other
    methods usually only limit the number of threads that are _used_,
    not the number of threads that are created.

    Parameters
    ----------
    n : int
        The maximum number of threads
    set_numba : bool, optional
        Set the :code:`'NUMBA_NUM_THREADS'` environment variable. If None, determine if
        Numba has been initialized and only set it if not.
        Numba may throw a :class:`RuntimeError` if the environment is
        altered after the number of threads has been set already.
    """
    if set_numba is None:
        from numba.np.ufunc import parallel
        set_numba = not getattr(parallel, "_is_initialized", True)
    try:
        env_keys = [
            "MKL_NUM_THREADS", "OMP_NUM_THREADS",
            "OPENBLAS_NUM_THREADS"
        ]
        if set_numba:
            env_keys.append("NUMBA_NUM_THREADS")
        old_env = {
            k: os.environ[k]
            for k in env_keys
            if k in os.environ
        }
        os.environ.update({
            k: str(n)
            for k in env_keys
        })
        yield
    finally:
        os.environ.update(old_env)
        for key in env_keys:
            if key not in old_env:
                del os.environ[key]


@contextmanager
def mitigations():
    '''
    Enable known work-arounds to run in a threaded executor.

    * Disable pyFFTW interface cache:
      https://github.com/LiberTEM/LiberTEM-blobfinder/issues/35
    '''
    global __has_pyfftw
    global __has_pyfftw_threads
    global __mitigation_counter
    global __pyfftw_cached

    detect_pyfftw()

    try:
        with lock:
            if __mitigation_counter == 0:
                if __has_pyfftw:
                    import pyfftw  # noqa:F811
                    __pyfftw_cached = pyfftw.interfaces.cache.is_enabled()
                    pyfftw.interfaces.cache.disable()
            __mitigation_counter += 1
        yield
    finally:
        with lock:
            __mitigation_counter -= 1
            if __mitigation_counter == 0:
                if __has_pyfftw and __pyfftw_cached is not None:
                    import pyfftw  # noqa:F811
                    if __pyfftw_cached:
                        pyfftw.interfaces.cache.enable()
                    else:
                        pyfftw.interfaces.cache.disable()
                    __pyfftw_cached = None



from .base import BaseAnalysis, AnalysisResult, AnalysisResultSet
import libertem.udf.crystallinity as crystal
from .helper import GeneratorHelper


class FFTMaskTemplate(GeneratorHelper):

    short_name = "fft"

    def __init__(self, params):
        self.params = params

    def get_dependency(self):
        return ["from libertem.analysis import ApplyFFTMask"]

    # FIXME write and include documentation
    def get_docs(self):
        docs = ["# FFT Analysis"]
        return '\n'.join(docs)

    def get_analysis(self):
        temp_analysis = [f"fft_analysis = ApplyFFTMask(dataset=ds, parameters={self.params})",
                         "fft_result = ctx.run(fft_analysis, progress=True)"]
        return '\n'.join(temp_analysis)

    def get_plot(self):
        plot = ["plt.figure()",
                "plt.imshow(fft_result.intensity)",
                "plt.colorbar()"]
        return ['\n'.join(plot)]


class ApplyFFTMask(BaseAnalysis, id_="APPLY_FFT_MASK"):
    TYPE = "UDF"

    def get_udf(self):
        rad_in = self.parameters["rad_in"]
        rad_out = self.parameters["rad_out"]
        real_center = (self.parameters["real_centery"], self.parameters["real_centerx"])
        real_rad = self.parameters["real_rad"]
        return crystal.CrystallinityUDF(
            rad_in=rad_in, rad_out=rad_out,
            real_center=real_center, real_rad=real_rad,
        )

    def get_udf_results(self, udf_results, roi, damage):
        from libertem.viz import visualize_simple
        data = udf_results['intensity'].data
        return AnalysisResultSet([
            AnalysisResult(
                raw_data=data,
                visualized=visualize_simple(
                    data.reshape(self.dataset.shape.nav),
                    damage=True
                ),
                key="intensity", title="intensity",
                desc="result from integration over mask in Fourier space"
            ),
        ])

    @classmethod
    def get_template_helper(cls):
        return FFTMaskTemplate



import numpy as np

import libertem.udf.FEM as FEM
from libertem.analysis.fem import FEMAnalysis

from libertem.io.dataset.memory import MemoryDataSet


def test_smoke(lt_ctx):
    """
    just check if the analysis runs without throwing exceptions:
    """
    data = np.zeros([3*3, 8, 8]).astype(np.float32)
    dataset = MemoryDataSet(data=data, tileshape=(1, 8, 8),
                            num_partitions=2, sig_dims=2)
    result = FEM.run_fem(ctx=lt_ctx, dataset=dataset, center=(1, 1), rad_in=0, rad_out=1)
    assert np.allclose(result['intensity'].data, np.zeros(data.shape[0]))


def test_different_size(lt_ctx):
    """
    just check if the analysis runs without throwing exceptions:
    """
    data = np.zeros([3*3, 3, 3]).astype(np.float32) + [[0, 1, 0], [3, 0, 3], [0, 1, 0]]
    data[8] = data[8] + [[0, 0, 0], [0, 0, 0], [0, 0, 9]]
    dataset = MemoryDataSet(data=data, tileshape=(1, 3, 3),
                            num_partitions=2, sig_dims=2)
    result = FEM.run_fem(ctx=lt_ctx, dataset=dataset, center=(1, 1), rad_in=0, rad_out=1)
    assert np.allclose(result['intensity'].data, np.ones(data.shape[0]))


def test_fem_analysis(lt_ctx):
    data = np.zeros([3*3, 8, 8]).astype(np.float32)
    dataset = MemoryDataSet(data=data, tileshape=(1, 8, 8),
                            num_partitions=2, sig_dims=2)
    fem_analysis = FEMAnalysis(parameters={
        'cx': 1,
        'cy': 1,
        'ri': 0,
        'ro': 1,
    }, dataset=dataset)

    result = lt_ctx.run(fem_analysis)
    assert np.allclose(result['intensity'].raw_data, np.zeros(data.shape[0]))
